;; Programa: Package con agentes jugadores de PlanTron contra 
;;           generadores de plantas carnivoras basados en
;;           búsquedas con sistemas de producción
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: 3 (Sistemas de Producción II)
;; Autor:    aluXXXX ([Indica aquí tu nombre completo])
;; Fecha:    21/05/2008
;; Comentarios:
;;
;;           Este código supone un esqueleto de ejemplo de el
;;           fichero que ha de contener los agentes jugadores
;;           de PlanTron que el usuario defina y usen búsqueda
;;           en sistemas de producción.
;;
;;           Para personalizar este fichero ha de renombrarse,
;;           sustituyendo "aluXXXX" por el lo que desee. Además
;;           en este código ha de reemplazarse dondequiera que
;;           apareza "aluXXXX" o "ALUXXXX" por lo mismo, ya que
;;           el nombre del fichero sin extensión y el nombre
;;           del package aquí incluido han de coincidir (no
;;           sensitivo a mayúsculas).
;;
;;
;;           Para añadir más agentes habrán de generarse las
;;           funciones correspondientes para los agentes e 
;;           incluirlas como sublistas en el contenido de la
;;           constante (DEFCONSTANT) AGENTES. Para más información
;;           consultar ejemplos de "pt-agentes-simples.lisp" y 
;;           la documentación (DESCRIBE) de las constantes, variables
;;           Y funciones allí utilizadas.
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT, PT-BASE y ENVIO.
;;
;;           Las funciones accesibles más importantes son:
;;
;;             ALUXXXX-CAMINO-BA: Agente PlanTron que busca
;;             el camino a lugar lejos de plantas mendiante
;;             búsqueda en anchura
;;
;;             ALUXXXX-CAMINO-A: Agente PlanTron que busca
;;             el camino a lugar lejos de plantas mendiante
;;             búsqueda en A*
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;;           Hace uso del paquete PT-UTIL con agentes pasivos
;;           con distintas utilidades para asistir a agentes jugadores
;;           de PlanTron (agentes pasivos CONTADOR y PLANTA-CERCA) e
;;           interfaces para acceso a paquetes SHELL, GENERIC-SEACH y
;;           A_STAR.
;;
;;           Hace uso del paquete interfaz para la creación de sistemas
;;           de producción simples (SHELL) a través de PT-UTIL.
;;
;;           Hace uso del paquete marco genérico para la búsquedas en
;;           sistemas de producción simples (GENERIC-SEARCH) a través
;;           de PT-UTIL.
;;
;;           Hace uso del paquete para la búsquedas A* (A_STAR) en
;;           sistemas de producción simples a través de PT-UTIL
;;
;;           Hace uso de la función AGENTE-SIN-MIEDO del paquete
;;           PT-AGENTES-SIMPLES.
;;
;;           Hace uso inderecto del paquete PT-VIS mediante la
;;           solicitud de impresión del camino encontrado en la
;;           función PRINT-CAMINO.
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (21/05/2008): Primera versión
;;
;;
;;
;; Por hacer:
;;
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Cuando esté encerrado mover con otro agente y evitar
;;             seguir haciendo búsquedas en futuros turnos
;;               Dificultad: Muy fácil; Mejora esperada: Poca
;;           - Llegado al límite de tiempo usar camino a lugar
;;             encontrado mas alejado de planta
;;               Dificultad: Fácil; Mejora esperada: Media a alta
;;           - Cuando está muy próximo a planta mover con otro agente
;;               Dificultad: Fácil; Mejora esperada: Poca
;;           - Que *dist-obj* aumente o disminuya, según encuentre
;;             o no caminos (se autoadapte)
;;               Dificultad: Fácil;
;;               Mejora esperada: Media a alta
;;           - Usar otra estrategias heurística de búsqueda
;;               Dificultad: Media; Mejora esperada: Media
;;           - Usar tratamiento de estados visitados
;;               Dificultad: Depende, media o difícil;
;;               Mejora esperada: Media a alta
;;           - Preferir caminos que pasen menos cerca de plantas
;;               Dificultad: Depende, media o difícil;
;;               Mejora esperada: Media
;;           - Utilizar información de varias plantas cercanas
;;               Dificultad: Depende, media a muy difícil;
;;               Mejora esperada: Media
;;           - Reutilizar base de hechos de búsquedas anteriores
;;               Dificultad: Depende, media a muy difícil
;;               Mejora esperada: Media a alta
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "AGENTES-ALUXXXX-CAMINO"
  (:use "COMMON-LISP")
  (:documentation "Agentes jugadores de PlanTron basado en búsquedas.")
  (:export "AGENTES" "VERSION" "*DIST-OBJ*" "AGENTE-ALUXXXX-CAMINO-INI" "AGENTE-ALUXXXX-CAMINO-BA-INI" "AGENTE-ALUXXXX-CAMINO-BA" "AGENTE-ALUXXXX-CAMINO-A-INI" "AGENTE-ALUXXXX-CAMINO-A"))

(in-package "AGENTES-ALUXXXX-CAMINO")


;;
;; Constante versión
;; -----------------

(defconstant version "0.1"
  "Versión del package AGENTES-ALUXXXX-CAMINO.")


;;
;; Constante con lista de agentes presentes en este package
;; --------------------------------------------------------

(defconstant agentes
  '((:turno aluXXXX-camino-ba
     :inicio aluXXXX-camino-ba-ini
     :use ("CONTADOR" "PLANTA-CERCA" "SHELL" "GENERIC-SEARCH"))
    (:turno aluXXXX-camino-a
     :inicio aluXXXX-camino-a-ini
     :use ("CONTADOR" "PLANTA-CERCA" "SHELL" "GENERIC-SEARCH" "A_STAR")))
  "Agentes basados en búsquedas para jugar a PlanTron.")


;;
;; Definición de estructura de estados
;; -----------------------------------

(defstruct estado
  "Estructura para almacenar los estados del sistema de producción"
  (camino nil)       ; Camino como lista inversa de posiciones (x y)
  (lon-camino 0)     ; Longitud del camino
  (f-heur 0)         ; Valor de la función heurística f'
  (dir-ini nil))     ; Dirección inicial del camino: [nil, 0, 1, 2 ó 3]


;;
;; Definición de varialbes globales
;; --------------------------------

(defvar *f-heur* nil
  "Función heurística que evalua un estado en acción de regla.")

(defvar *dist-obj* 6
  "Distancia a la que debe estar el agente de la planta mas cercana para cumplir su objetivo")


;;
;; Funciones de apoyo para elaboración de reglas
;; ---------------------------------------------

(defun pos-posible-p (pos &optional dir)
  "Mira si es posible colocarse (sin planta, obstáculo o rastro de agente con que colisionar) en la posición resultante de moverse en la dirección DIR a la posición POS. Devuelve T en tal caso y NIL en caso contrario. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. DIR ha de ser un número entero (dirección) entre 0 y 3 (norte, este, sur y oeste respectivamente) o NIL si se quiere consultar la posición POS."
  (not (envio:consulta-tipo (envio:consulta-pos-tab pos dir))))

(defconstant dir-inc
  (make-array '(4 2)
	      :initial-contents
	      '((1 -1)    ; Norte
		(0 +1)    ; Este
		(1 +1)    ; Sur
		(0 -1)))  ; Oeste
  "Array para transformar de dirección a número de coordenada a modificar y valor de incremento.")

(defmacro dir2dim-inc (dir)
  `(aref dir-inc ,dir 0))

(defmacro dir2val-inc (dir)
  `(aref dir-inc ,dir 1))

(defun pos-inc-dir (pos dir)
  "Calcula nueva lista con la posición resultante de desplazarse en la dirección DIR respecto a POS."
  (let ((pos-inc (copy-list pos)))
    (setf (nth (dir2dim-inc dir) pos-inc)
	  (+ (nth (dir2dim-inc dir) pos-inc)
	     (dir2val-inc dir)))
    pos-inc))

(defun dist-planta-cerca (pos)
  "Calcula la distancia city-block desde la posición POS a la planta más cercana a dicha posición. Si no hay plantas devuelve NIL."
  (let ((pc (pt-util:planta-cerca pos)))
    (if (consp pc)
	(second (first pc)))))


;;
;; Funciones para generar sistema de producción
;; --------------------------------------------

(defun estados-iniciales (id)
  "Genera lista de estados iniciales del jugador con identificador ID."
  (let ((nue-est (make-estado :camino (list (envio:consulta-pos-jug id)))))
    (if *f-heur*
	(setf (estado-f-heur nue-est) (funcall *f-heur* nue-est)))
    (list nue-est)))

(defun estado-objetivo-p (est)
  "Chequea si un estado EST es estado objetivo o NIL."
  ; Si se ha acabado el tiempo o no se ha encontrado solución
  (or (pt-util:consulta-contador) (null est)
      (let ((dt (dist-planta-cerca (first (estado-camino est)))))
	(or (not dt)              ; o no hay plantas
	    (>= dt                ; o se ha alcanzado la distancia 
		*dist-obj*)))))   ; *dist-obj* a la planta mas cercana

(defun precond-regla-dir (est dir)
  "Generador de precondicion de regla de mover el estado EST en dirección DIR."
  (let ((pos-est (first (estado-camino est))))
    ; La posición no pertenece al propio camino
    (and (not (member (pos-inc-dir pos-est dir)
		      (rest (estado-camino est))
		      :test #'equal))
	 ; Y la posición se puede pisar
	 (pos-posible-p pos-est dir))))

(defun accion-regla-dir (est dir)
  "Generador de acción de regla de mover el estado EST en dirección DIR."
  (let ((nue-est (make-estado :camino 
			      (cons (pos-inc-dir (first (estado-camino est)) dir)
				    (estado-camino est))
			      :lon-camino (1+ (estado-lon-camino est))
			      :dir-ini (or (estado-dir-ini est) dir))))
    ; Asignar valor heurísticos, si se requiere
    (if *f-heur*
	(setf (estado-f-heur nue-est) (funcall *f-heur* nue-est)))
    nue-est))


;;
;; Funciones de apoyo para agentes de búsqueda de camino
;; -----------------------------------------------------

(defun print-camino (camino)
  "Solicita imprimir la lista de posiciones dada en CAMNIO a los agentes capaces de acpetar solicitudes tipo :punteo."
  (envio:broadcast :punteo (reverse camino)))


;;
;; Inicializador de agentes de búsqueda de camino
;; ----------------------------------------------

(defun aluXXXX-camino-ini (id tipo-juego)
  "Inicialización genérica de agente que busca camino."
  (declare (ignore id tipo-juego))
  ; Eliminar las reglas existentes
  (pt-util:limpia-reglas)
  ; Insertar las nuevas reglas
  (pt-util:regla #'(lambda (est) (precond-regla-dir est 3))  ; Regla Oeste
	 #'(lambda (est) (accion-regla-dir est 3)))
  (pt-util:regla #'(lambda (est) (precond-regla-dir est 1))  ; Regla Este
	 #'(lambda (est) (accion-regla-dir est 1)))
  (pt-util:regla #'(lambda (est) (precond-regla-dir est 0))  ; Regla Norte
	 #'(lambda (est) (accion-regla-dir est 0)))
  (pt-util:regla #'(lambda (est) (precond-regla-dir est 2))  ; Regla Sur
	 #'(lambda (est) (accion-regla-dir est 2)))
  ; Cargar heurística a utilizar (sin heurística)
  (setf *f-heur* nil)
  t)


;;
;; Agente búsqueda de camino en anchura
;; ------------------------------------

(defun aluXXXX-camino-ba-ini (id tipo-juego)
  "Inicialización de agente que busca camino en anchura."
  ; Llamar a inicialización de agentes camino
  (aluXXXX-camino-ini id tipo-juego))

(defun aluXXXX-camino-ba (id)
  "Agente jugador PlanTron que busca camino a lugar lejos de plantas moviéndose en anchura, si se detecta que se ha quedado encerrado se mueve como el agente-sin-miedo."
  ; Iniciar contador de tiempo
  (pt-util:inicia-contador)
  ; Lanzar búsqueda
  (let ((est-sol
	 (pt-util:generic-search
	  (estados-iniciales id)
	  #'estado-objetivo-p
	  #'pt-util:hijos
	  :display-fn #'null
	  :merge-fn   #'(lambda (new-states)
			  (nconc pt-util:*search-queue* new-states))
	  :dequeue-fn #'(lambda () (pop pt-util:*search-queue*)))))
    ; Preguntar por condiciones de parada
    (cond
     ((null est-sol)              ; No encontrada solución (encerrado)
      (envio:mensaje "Encerrado :( lanzo sin-miedo")
      (pt-agentes-simples::agente-sin-miedo id))  ; Como agente-sin-miedo
     
     ((pt-util:consulta-contador) ; Agotado el tiempo de búsqueda
      (envio:mensaje "Agotado el tiempo de búsqueda")
      (estado-dir-ini est-sol))                ; Mover según primero frente
     
     (t                           ; Encontrado camino
      (envio:mensaje "Encontrado camino anchura")
      (print-camino (estado-camino est-sol))    ; Imprimir camino hallado
      (estado-dir-ini est-sol)))))              ; Mover según primero frente


;;
;; Agente búsqueda de camino A*
;; ----------------------------

(defun f-dist-planta (est)
  "Función heurística subdividida con coste del camino acumulado + distancia city block que falta para alejarse hasta *dist-obj* de la planta mas cercana"
  (let ((dt (dist-planta-cerca (first (estado-camino est)))))
    (+ (estado-lon-camino est)
       (if dt              ; Si localiza planta
	   (- *dist-obj*   ;   estima la distancia hasta separarse *dist-obj*
	      dt)          ; sino
	 0))))             ;   supone ya separado 

(defun estado-objetivo-a-p (est)
  "Chequea si un estado EST es estado objetivo o NIL. Aprovecha el campo F-HEUR del estado."
  ; Si se ha acabado el tiempo o no se ha encontrado solución
  (or (pt-util:consulta-contador) (null est)
      (>= (estado-lon-camino est) (estado-f-heur est))))  ; o ya se ha separado

(defun igual-estado (est1 est2)
  "Comprueba si el estado EST1 es igual al EST2"
  ; Considera estados iguales si su camino acaba en la misma posición
  (equal (first (estado-camino est1))
         (first (estado-camino est2))))

(defun aluXXXX-camino-a-ini (id tipo-juego)
  "Inicialización de agente que busca camino con A*."
  ; Llamar a inicialización de agentes camino
  (aluXXXX-camino-ini id tipo-juego)
  ; Inicializar variables para heurísticas del A*
  (setf *f-heur* #'f-dist-planta)                     ; Fun. f' para reglas
  (setf pt-util:*coste-acum-fn* #'estado-lon-camino)  ; Fun. g  para búsqueda
  (setf pt-util:*eval-fn* #'estado-f-heur)            ; Fun. f' para búsqueda
  (setf pt-util:*igual-est-fn* #'igual-estado)        ; Pred. estados iguales
  t)

(defun aluXXXX-camino-a (id)
  "Agente jugador PlanTron que busca camino a lugar lejos de plantas moviéndose con algoritmo A, si se detecta que se ha quedado encerrado se mueve como el agente-sin miedo."
  ; Iniciar contador de tiempo
  (pt-util:inicia-contador)
  ; Lanzar búsqueda
  (let ((est-sol
	 (pt-util:generic-search
	  (estados-iniciales id)
	  #'estado-objetivo-a-p
	  #'pt-util:hijos
	  :display-fn #'null
	  :merge-fn   #'pt-util:priority-a-merge-fn
	  :dequeue-fn #'pt-util:priority-a-dequeue-fn)))
    ; Preguntar por condiciones de parada
    (cond
     ((null est-sol)              ; No encontrada solución (encerrado)
      (envio:mensaje "Encerrado :( lanzo sin-miedo")
      (pt-agentes-simples::agente-sin-miedo id))  ; Como agente-sin-miedo
     
     ((pt-util:consulta-contador) ; Agotado el tiempo de búsqueda
      (envio:mensaje "Agotado el tiempo de búsqueda")
      (estado-dir-ini est-sol))                ; Mover según primero frente
     
     (t                           ; Encontrado camino
      (envio:mensaje "Encontrado camino A*")
      (print-camino (estado-camino est-sol))    ; Imprimir camino hallado
      (estado-dir-ini est-sol)))))              ; Mover según primero frente

