;; --------------------------------
;; Presentación de entorno PlanTron
;; --------------------------------
;;  Este fichero recoge una serie de instrucciones
;;  que guían el aprendizaje del entorno de agentes
;;  PlanTron (ver fichero README).
;;
;;  Puedes ir copiando sobre una consola las
;;  instrucciones aquí descritas.
;;


;; Para cargar el entorno
;; ----------------------

;; Arrancamos en consola el CMUCL lisp con 'cmucl'
#|

cmucl

|#
;; También es posible usar SBCL lisp


;; Cargamos el entorno en el lugar en que esté,
;; en el CC de la ETSII se encuentra en el
;; directorio /scratch/plantron
;;

(load "/scratch/plantron/pt-loader")

;; ¡Atención! Cargar siempre desde scratch, ya
;; que pueden haber modificaciones en los códigos.
;; Comprobar versión contenido fichero VERSION o
;; ejecutar en ahora (version-pt)



;; Partida con humano
;; ------------------

;; Movimientos de teclas (teclado numérico):
;;   8: arriba, 2:abajo, 6: derecha, 4:izquierda

;; ¡Atención! Asegurarse antes de jugar que el
;; tamaño de la terminal es al menos de 80x24
;; (VT100)

(pt:partida)


;; La expresión anterior es equivalente a:

(pt:partida "AGENTE-HUMANO")


;; Descripción de funciones y variables con describe

(describe 'pt:partida)


;; Información del agente

(pt::info-jug "AGENTE-HUMANO")



;; Partida con agente mitad
;; ------------------------

(pt:partida "AGENTE-MITAD")

;; Más rápido, opción :sleep-cs <milisegundos/turno>

(pt:partida "AGENTE-MITAD" :sleep-cs 1)

;; Utilización permanente de :sleep-cs

(setf pt-base:*sleep-cs* 15)

;; Generar plantas en modo menos virulento,
;; variable pt-agentes-genplc:*var-sigm-hacia*

(setf pt-agentes-genplc:*var-sigm-hacia* 4)

(pt:partida "AGENTE-MITAD" :sleep-cs 1)


;; Generar plantas en modo mas virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 128)

(describe 'pt-agentes-genplc:*var-sigm-hacia*)

(pt:partida "AGENTE-MITAD" :sleep-cs 3)



;;
;; Consultar jugadores disponibles
;; -------------------------------

(pt:jug)

(describe 'pt:jug)

(pt:jug :full t)



;; Partida con agente lados
;; ------------------------

;; Generar plantas en modo menos virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 4)

(pt:partida "AGENTE-LADOS" :sleep-cs 1)

;; Generar plantas en modo mas virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 128)

(pt:partida "AGENTE-LADOS" :sleep-cs 5)



;; Partida con agente espiral
;; --------------------------

;; Generar plantas en modo menos virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 4)

(pt:partida "AGENTE-ESPIRAL" :sleep-cs 1)

;; Generar plantas en modo mas virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 128)

(pt:partida "AGENTE-ESPIRAL" :sleep-cs 5)



;; Partida con agente muralla
;; --------------------------

(pt:partida "AGENTE-MURALLA" :sleep-cs 1)



;; Partida con agente sin-miedo
;; ----------------------------

;; Generar plantas en modo menos virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 4)

(pt:partida "AGENTE-SIN-MIEDO" :sleep-cs 5)

;; Generar plantas en modo mas virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 128)

(pt:partida "AGENTE-SIN-MIEDO" :sleep-cs 1)



;; Partida con agente cobarde
;; --------------------------

;; Generar plantas en modo menos virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 4)

(pt:partida "AGENTE-COBARDE" :sleep-cs 5)

;; Generar plantas en modo mas virulento

(setf pt-agentes-genplc:*var-sigm-hacia* 128)

(pt:partida "AGENTE-COBARDE" :sleep-cs 5)



;; Partida con agente hormiga borracha
;; -----------------------------------

(pt:partida "AGENTE-HORMIGA-BORRACHA" :sleep-cs 10)



;; Cargar y ejecutar un nuevo agente
;; ---------------------------------

(pt:load-ag "agentes-aluXXXX.lisp")

(describe 'pt:load-ag)

(pt:jug)

(pt:partida "ALUXXXX-MITAD" :sleep-cs 1)



;; Partidas con dos agentes
;; ------------------------

(pt:partida "AGENTE-MITAD" "AGENTE-ESPIRAL" :sleep-cs 1)


(pt:partida "AGENTE-MITAD" "AGENTE-MURALLA" :sleep-cs 1)


(pt:partida "AGENTE-MURALLA" "AGENTE-SIN-MIEDO" :sleep-cs 1)


(pt:partida "AGENTE-SIN-MIEDO" "AGENTE-COBARDE" :sleep-cs 1)



;; Partida con trazas de consultas, opción :vis <nivel>
;; ----------------------------------------------------

(pt:partida "AGENTE-COBARDE" :sleep-cs 5 :vis 2)

(pt:partida "AGENTE-COBARDE" :sleep-cs 1 :vis 2)


;; Sin visualizador

(pt:partida "AGENTE-COBARDE" :sleep-cs 1 :vis nil)


;; Utilización permanente de un visualizador

(setf pt:*vis* 2)

(pt:partida "AGENTE-COBARDE" :sleep-cs 1)



;; Partida con otros generadores de plantas, opción :gen-pla
;; --------------------------------------------------------

;; Generador de plantas aleatorio

(pt:partida "AGENTE-COBARDE" :sleep-cs 1 :gen-pla 2 :vis 2)


;; Generador de plantas en cúmulos no incremental,
;; detener con Ctrl-C y comando TOP

(pt:partida "AGENTE-MURALLA" :sleep-cs 1 :gen-pla 1 :vis 2)



;; Pintar el tablero en otra terminal, opción :dev-vis
;; ---------------------------------------------------

;; Localizar antes con comando 'w' el dispositivo
;; asociado a otra terminal existente,
;; por ejemplo para "/dev/pts/2"

(pt:partida "AGENTE-ESPIRAL" :sleep-cs 1 :dev-vis "/dev/pts/2")


;; Traceamos función dir-posible-pref

(trace pt-agentes-simples::dir-posible-pref)

(pt:partida "AGENTE-ESPIRAL" :sleep-cs 1 :dev-vis "/dev/pts/2")

(untrace  pt-agentes-simples::dir-posible-pref)


;; Utilización permanente de otra terminal

(setf pt:*dev-vis* "/dev/pts/2")

(pt:partida "AGENTE-ESPIRAL" :sleep-cs 1)



;; Partidas paso a paso
;; --------------------

(setf pt-vis:*step* t)

(pt:partida "AGENTE-ESPIRAL" :vis 2 :dev-vis "/dev/pts/2")

(pt:partida "AGENTE-COBARDE" :vis 2 :dev-vis "/dev/pts/2")

(describe 'pt-vis:*step*)



;; Partidas de todos los agentes no interactivos
;; ---------------------------------------------

(pt:partida-todos :sleep-cs 5)

(describe 'pt:partida-todos)



;; Resultados de 5 partidas del agente espiral
;; ejecutadas sin visualizador

(pt:partida-todos '("AGENTE-ESPIRAL") :num 5 :vis nil :sleep-cs 1)



;; Envío de agentes: trabajo con repositorio arena
;; -----------------------------------------------

;; Mostrar los agentes en repositorio arena

(pt-arena:ls)

(describe 'pt-arena:ls)

(pt-arena:ls "*" t)


;; Chequear y enviar agentes al repositorio arena
;; ¡Atención! Asegurarse de ejecutar esta instrucción
;; desde la máquina tonga (exthost) y cargar siempre 
;; el entorno desde scratch (/scratch/plantron/pt-loader)

(pt-arena:send "agentes-aluXXXX.lisp")

;; Chequear cuidadosamente los mensajes emitidos y asegurarse
;; de que su agente aparece listado en el entorno arena

(pt-arena:ls)


;; Cargar agente del repositorio arena

(pt-arena:load-ag "agentes-aluXXXX")


;; Eliminar un agente del repositorio arena
;; (no anula el envío)

(pt-arena:del "agentes-aluXXXX")

(describe 'pt-arena:del)


;; Tablas de resultados

(pt-arena:hof)

(describe 'pt-arena:hof)



;; Ejemplo de agentes buscadores de caminos
;; ----------------------------------------

; Cargar agentes de caminos del repositorio arena

(pt-arena:load-ag "agentes-aluXXXX-camino")

(describe 'pt-arena:load-ag)


; Lanzar agente búsqueda de camino en anchura

(pt:partida "ALUXXXX-CAMINO-BA")

(pt:partida "ALUXXXX-CAMINO-BA" :vis 2)


; Lanzar agente búsqueda de camino en A*

(pt:partida "ALUXXXX-CAMINO-A" :vis 2)

(pt:partida "ALUXXXX-CAMINO-A" :vis 2 :sleep-cs 60)



;; Compilación de agentes
;; ----------------------

(compile-file "agentes-aluXXXX-camino.lisp")

(pt:load-ag "agentes-aluXXXX-camino.x86f")



;; Salir del cmucl lisp
;; --------------------

(quit)
