;; Programa: Package con agentes jugadores simples de PlanTron
;;           contra generadores de plantas carnivoras
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (lisp)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    28/04/2008
;; Comentarios:
;;
;;           Incluye agentes para jugar al PlanTron simples, esto
;;           es, agentes que no tiene memoria y su comportamiento
;;           viene guiado en función del entorno que perciben en
;;           cada instante (excepto el agente-hormiga-borracha,
;;           que hace uso del generador de números aleatorios).
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT-BASE y ENVIO.
;;
;;           Los agentes incluidos son, ordenados según la
;;           complejidad de su código, los siguientes:
;;
;;             AGENTE-MITAD
;;             AGENTE-LADOS
;;             AGENTE-ESPIRAL
;;             AGENTE-MURALLA
;;             AGENTE-SIN-MIEDO
;;             AGENTE-COBARDE
;;             AGENTE-HORMIGA-BORRACHA
;;
;;           Ver la documentación existente para cada una de ellos.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (28/04/2008): Primera versión.
;;
;; Por hacer:
;;
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "PT-AGENTES-SIMPLES"
  (:nicknames "PT-A-S")
  (:use "COMMON-LISP")
  (:export "AGENTES" "VERSION" "DIR-PLANTA-P" "DIR-POSIBLE-P" "DIR-INC" "DIR-POSIBLE-PREF")
  (:documentation "Agentes para jugar al PlanTron simples, esto es, agentes que no tiene memoria y su comportamiento viene guiado en función del entorno que perciben en cada instante (excepto el agente-hormiga-borracha, que hace uso del generador de números aleatorios)."))


(in-package "PT-AGENTES-SIMPLES")


;;
;; Constante con lista de agentes presentes en este package
;; --------------------------------------------------------

(defconstant agentes
  '((:turno agente-mitad)
    (:turno agente-lados)
    (:turno agente-espiral)
    (:turno agente-muralla)
    (:turno agente-sin-miedo)
    (:turno agente-cobarde)
    (:turno agente-hormiga-borracha))
  "Agentes simples para jugar a PlanTron.")


;;
;; Constante versión
;; -----------------

(defconstant version "0.1"
  "Versión del package PT-AGENTES-SIMPLES.")


;;
;; Funciones de tratamiento de direcciones y posiciones
;; ----------------------------------------------------

(defun dir-planta-p (pos dir)
  "Mira si han una planta en la posición resultante de moverse en la dirección DIR a la posición POS. Devuelve DIR en tal caso y NIL en caso contrario. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. DIR ha de ser un número entero (dirección) entre 0 y 3 (norte, este, sur y oeste respectivamente)."
  (if (eq (envio:consulta-tipo (envio:consulta-pos-tab pos dir)) 'envio:planta)
      dir))

(defun dir-posible-p (pos dir)
  "Mira si es posible colocarse (sin obstáculo o rastro de agente con que colisionar) en la posición resultante de moverse en la dirección DIR a la posición POS. Devuelve DIR en tal caso y NIL en caso contrario. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. DIR ha de ser un número entero (dirección) entre 0 y 3 (norte, este, sur y oeste respectivamente)."
  (if (not (envio:consulta-tipo (envio:consulta-pos-tab pos dir)))
      dir))

(defun number-o-0 (elm)
  "Si ELM es un número lo devuelve, sino devuelve 0."
  (if (numberp elm)
      elm
    0))

(defun dir-inc (dir dir-ini)
  "Calcula la dirección DIR relativa a DIR-INI, esto es, gira DIR-INI pasos en sentido agujas del reloj sobre DIR. DIR y DIR-INI han de ser números enteros (direcciones,  numero entero entre 0 y 3: norte, este, sur y oeste respectivamente), si no se consideran cero (norte)."
  (rem (+ (number-o-0 dir) (number-o-0 dir-ini)) 4))

(defun dir-posible-pref (pos lis-dir-pref &optional dir-ini)
  "Busca la primera dirección posible (ver función DIR-POSIBLE-P) alrededor de POS siguiendo el orden de LIS-DIR-PREF, si DIR-INI está instanciado las direcciones de LIS-DIR-PREF son relativas a DIR-INI. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. LIS-DIR-PREF ha de ser una lista con números enteros (direcciones). DIR-INI ha de ser un número entero (dirección)."
  (if (consp lis-dir-pref)
      (or (dir-posible-p pos (dir-inc (first lis-dir-pref) dir-ini))
	  (dir-posible-pref pos (rest lis-dir-pref) dir-ini))))


;;
;; Agente NSEO
;; -----------

(defun agente-mitad (id)
  "Agente jugador PlanTron que sólo busca la mitad."
  (dir-posible-pref (envio:consulta-pos-jug id) '(0 2 1 3)))


;;
;; Agente NESO
;; -----------

(defun agente-lados (id)
  "Agente jugador PlanTron que se va por los lados."
  (dir-posible-pref (envio:consulta-pos-jug id) '(0 1 2 3)))


;;
;; Agente espiral derecha
;; ----------------------
  
(defun agente-espiral (id)
  "Agente jugador PlanTron que da vueltas en espiral desenrrollándose!"
  (dir-posible-pref (envio:consulta-pos-jug id)
		    '(1 0 3)
		    (first (envio:consulta-est-jug id))))


;;
;; Agente muralla
;; --------------

(defun agente-muralla (id)
  "Agente jugador PlanTron que va temerariamente enrrollándose!"
  (dir-posible-pref (envio:consulta-pos-jug id)
		    '(0 1 3)
		    (first (envio:consulta-est-jug id))))


;;
;; Agente sin-miedo
;; ----------------

(defun alejarse-p (pos lis-dir-pref)
  (cond
   ((endp lis-dir-pref) nil)
   ((and (dir-planta-p pos (dir-inc (first  lis-dir-pref) 2))
	 (dir-posible-p pos (first lis-dir-pref)))
    (first  lis-dir-pref))
   (t (alejarse-p pos (rest lis-dir-pref)))))

(defun agente-sin-miedo (id)
  "Agente jugador PlanTron que solo cuando se le acerca una planta carnivora trata de mover alejándose de ella con un paso en dirección contraria."
  (alejarse-p (envio:consulta-pos-jug id) '(0 2 1 3)))


;;
;; Agente cobarde
;; --------------

(defun huir-p (pos lis-dir-pref)
  (cond
   ((endp lis-dir-pref) nil)
   ((and (not (dir-posible-p pos (dir-inc (first  lis-dir-pref) 2)))
	 (dir-posible-p pos (first lis-dir-pref)))
    (first  lis-dir-pref))
   (t (huir-p pos (rest lis-dir-pref)))))

(defun agente-cobarde (id)
  "Agente jugador PlanTron que en cuanto se le acerca una planta carnivora trata de salir corriendo en dirección contraria."
  (huir-p (envio:consulta-pos-jug id) '(0 2 1 3)))


;;
;; Agente hormiga borracha
;; -----------------------

(defvar random-state-hormiga-borracha (make-random-state *random-state*)
  "Estado random para el AGENTE-HORMIGA-BORRACHA.")

(defun agente-hormiga-borracha (id)
  "Agente jugador PlanTron que da tumbos y se agarra a las farolas. Elige unmovimientro aleatoriamente entre los posibles (quedarse quieto y direcciones vacías a su alrededor)."
  (let* ((pos (envio:consulta-pos-jug id))
	 (lis-dir (cons nil (remove nil (mapcar (lambda (dir) (dir-posible-p pos dir))
						'(0 1 2 3))))))
    (nth (random (length lis-dir) random-state-hormiga-borracha)
	 lis-dir)))

