;; Programa: Package para entorno base para agentes de juego PlanTron
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    07/04/2006
;; Comentarios:
;;           Este package es el encargado de controlar el desarrollo
;;           de una partida PlanTron entre agentes jugadores.
;;           Almacena todo el estado del juego (tipo de partida, 
;;           tablero, posición de jugadores, dirección de movimiento
;;           de éstos, ...).
;;
;;           Se comunica con los agentes mediante un package
;;           que hace de interface de comunicaciones denominado
;;           ENVIO. De esta forma PT-BASE se mantiene
;;           independiente del sistema de comunicación empleado
;;           (llamadas a función, mensajes vía sockets, ...). 
;;
;;           Los tipos de partidas que puede controlar se definen
;;           mediante el parámetro TIPO pasado a la función
;;           INICIO, (ver función EXTRC-TIPO-PARTIDA). Éste almacena
;;           los siguientes datos:
;;
;;              - Número de jugadores (agentes)
;;              - Dimensiones del tablero (x y)
;;              - Sin-bordes o con bordes exteriores (t ó nil)
;;              - Centésimas seg. de retardo entre movimientos
;;              - No-verbose o verbose, indica la info. emitida
;;              - Juego síncrono (nil) o asíncrono (t), esto es,
;;                si se espera o no por movim. de agentes (hang)
;;              - Controlar el tiempo entre movimientos en servidor
;;                de juego (nil) o no (t) (para ejecuciones masivas).
;;                No se transmite a los agentes esté último.
;;
;;           Para representar el tablero se utilizan estructuras
;;           tipo PT (ver PT).
;;
;;           Las funciones accesibles más importantes, incluidas en
;;           "pt-base-package.lisp", permiten recibir mensajes
;;           originarios en los agentes mediante la interface
;;           de comunicación (ENVIO).
;;           Ver la documentación existente para cada una de ellas.
;;
;;
;; Dependencias:
;;           Hace uso de funciones del paquete ENVIO (buscar 
;;           "envio:") que lleva la comunicación del juego.
;;
;; Modificaciones:
;;
;;           - Ver 0.0 (07/04/2006): Comienzo de la codificación
;;
;;           - Ver 0.1 (11/04/2006): Versión inicial con código utlizable
;;             para gcl, aunque hay código para CMUCL y opciones para 
;;             insertar clisp. Realizado generador de estructura de
;;             problemas, funcciones de acceso/modificación, interface
;;             funcional básica con agentes, visualizador de partidas y
;;             agentes humano, rombo, espiral y aleatorio.
;;
;;           - Ver 0.2 (13/04/2006): División en dos paquetes: "tron-base"
;;             (tron-base.lsp) y "trom-comm" (tron-comm-fun.lsp)
;;             y diversos agentes (tron-vis-xterm.lsp, 
;;             tron-agentes-humanos.lsp y tron-agentes-simples.lsp)
;;             Uso de cargadores de paquetes tron-base-package.lsp y
;;             tron-comm-fun-package.lsp.
;;             Codificado para jugar sin bordes (salto dimensional).
;;
;;           - Ver 0.2.1 (29/04/2006): Reparado bug "Error en jugar
;;             sin bordes, no sale bien por arriba ni por la izquierda.
;;             Cheque de posiciones fuera devuelven obstáculo."
;;
;;           - Ver 0.2.2 (11/05/2006): Mejorada documentación.
;;
;;           - Ver 0.2.3 (30/05/2006): Añadido :sleep-nil en tipo-juego.
;;
;;           - Ver 0.3 (21/4/2007): Cambio de nombre de tron-base a 
;;             pt-base. Cambio extensión de "lsp" a "lisp". Cambio 
;;             también nombre de variables, funciones, etc de "tron" a 
;;             "pt".
;;
;;           - Ver 1.0 (1/05/2007): Ajustes para versión con nueva
;;             interface de comunicaciones en package "ENVIO"
;;
;;           - Ver 1.0.1 (8/05/2007): Retoques en documentación. Ajuste
;;             de *ratio-ene-jug-ini*
;;
;;           - Ver 1.1 (18/05/2007): Incorporados seguimientos :seg-npla
;;             y :seg-mpla
;;
;;           - Ver 1.1.1 (24/05/2007): Añadido posiblidad de keys
;;             de agentes no preestablecidas para prestar servicios
;;             con keywords-nofun-ag
;;
;;           - Ver 1.2 (22/04/2008): Añadido envios de :turno a
;;             jugadores tipo genpla. Posiblidad de plantas carnivoras
;;             que matan a gendirs, con pla-car en tipo de partida.
;;             Posiblidad de movimiento no continuo de generadores de
;;             dirección desactivando mov-con en tipo de partida.
;;             Reparado bug que generaba datos incorrectos para 
;;             seguimiento :seg-est. Permitir recoger datos de
;;             movimientos sincrona y asincronamente simultaneamente.
;;             Modificado get-accion para solo permitírlo en modo
;;             asíncrono
;;
;;           - Ver 1.2.1 (22/05/2008): Envío de avisos de cambio de
;;             posición de genplas antes de turno de gendirs.
;;             Separación de obtención de movimientos de genpla y
;;             de gendir.
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - Chekeos de consistencia de datos de entrada
;;           - Verificar funcionamiento de todo compilado
;;           - Evitar que se llame a seguimiento a agente si aún no
;;             se ha inicializado, por rejemplo con pt-vis
;;
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Probar todo bajo otros lisps
;;             (gcl, clisp linux clisp win, cmucl, ...)
;;           - Verificar posiblidad de jugar con /= 2 jugadores
;;           - Preparar modo de instalación
;;           - ¿simplificaría que jugador 1 = 0 y jugardor 2 = 1?
;;           - Distintos tipos de verbose: sin vis, vis pobre,
;;             media, rica
;;           - Ver que poner con verbose y que con visualizador
;;             de partida
;;           - Diferenciar verbose (logs) de tipo de visualizador
;;           - Calcular tiempos que tardan en decidir los agentes
;;           - Buscar implementación alternativa a contador
;;             (¿con eventos?)
;;           - ¿Identificador de partida que se transmita siempre?
;;           - ojo, el último movimiento no se informa ¿ok?
;;           - ¿Un query sobre el estado del juego?
;;           - ¿(setf *tablero* nil) al final de la partida?
;;           - ¿Campos de PT POS-JUG y DIR-JUG como matrices?
;;           - Modificar para admitir inicializaciones de más de 
;;             8 jugadores
;;           - Hacer el código reentrante
;;           - Juego variante Tron con tamaño de cola máximo,
;;             nueva tag :cola en tipo-juego.
;;           - Implementar :ES-COMIDO :ES-MATADO
;;           - Recoger desde :turno y procesar movimientos de
;;             jugadores genpla
;;           - Considerar usar campo ien-jug en estructura pt
;;             para almacenar incrementos y decrementos de energia
;;             con los movimientos para los jugadores
;;           - Añadir keyword de servicio :COME que avisa a un
;;             agente que ha comido algo
;;           - Permitir coexistir generadores de plantas carnívoras
;;             y no carnívoras
;;           - Permitir generadores de direcciones de distintos
;;             niveles: predadores en cadena
;;
;;
;;           - Mejoras en estilo de programación
;;             ---------------------------------
;;           - Estructurar disposición de código
;;           - Reescribir más claramente algunas partes del código...
;;           - ¿Cambiar números por símbolos en: direcciones, ...
;;           - ¿llamadas con envio: es lo mejor?
;;           - ¿Accesos impedidos a pt-mat pt-....? 
;;           - ¿Vale la pena la estructura de pt-dir-jug
;;             (ver check-set-inc-pos)?
;;           - ¿Meter en estructura pt tipo-juego?
;;           - Comentar mejor estructura TRON
;;           - Pasar keywords de keywords-fun-ag :seg-cposj 
;;             :seg-cestj :seg-cpost :rec-msj a ENVIO
;;           - Ver si eliminar definitivamente keys :servicio y
;;             :servicio-id
;;
;;
;;

(unless (find-package "PT-BASE")
  (load "pt-base-package"))

(in-package "PT-BASE")


;; Variables version
;; -----------------

(defconstant version "1.2.1" "Versión del package PT-BASE.")

(defun version ()
  "Devuelve la versión del paquete PT-BASE que se ha cargado."
  version)

(defvar version-entorno "`cat VERSION`"
  "Versión de todo el entorno PlanTron.")

(defvar *juego* "PlanTron"
  "Título del juego implementado.")


;; Constantes
;; ----------

(defconstant obs 0
  "Contiene el número que representa los obstáculos (#) dentro de la matiz del campo MAT de la estructura tipo PT.")

(defconstant tipos-jug
  '(nil obstaculo choque jugador planta desconocido) 
  "Lista de posibles valores que puede devolver la función TIPO-JUG.")

(defconstant tipos-ag
  '(nil gen-dir gen-pla pasivo)
  "Lista de posibles valores de tipos de agentes. El significado de cada una de los tipos es:
   GEN-DIR   : Agente del tipo generador de direcciones. Jugador que puede
               moverse en direcciones 0, 1, 2 y 3 correspondientes a Norte,
               Este, Sur y Oeste
   GEN-PLA   : Agente generador de plantas. Jugador que puede realizar
               peticiones para generar plantas (ENVIO:PET-GEN-PLANTA)
   PASIVO    : Agente pasivo, solo observa o puede proveer servicios a
               otros agentes, pero no generar direcciones o plantas
")

(defconstant keywords-fun-ag
  '(:inicio :turno :fin :es-comido :es-matado :seg-pos :seg-est :seg-npla :seg-mpla :seg-cposj :seg-cestj :seg-cpost :rec-msj :servicio :servicio-id)
  "Lista de posibles keywords de los distintos tipos de funciones que pueden tener los agentes para comunicarse con ENVIO. El significado de cada una de las keywords admisibles son:
  :INICIO     : Función a llamar en el inicio de la partida, puede existir para 
                cualquier tipo de agente. Recibe el identificador del agente
                llamado y el tipo de partida (ver EXTRC-TIPO-PARTIDA)
  :TURNO      : Función a llamar cada vez que le toca el turno de mover a un
                agente en partidas sícronas. Sólo utilizables por agentes
                activos: generadores de dirección (jugadores móviles) o
                generadores de plantas. Recibe el identificador del agente
                llamado. En el caso de los generadores de dirección el
                resultado devuelto ha de ser la dirección elegida para moverse
                (ver PT-BASE::DIR-JUGADOR) o NIL si se desea seguir en la misma
                dirección.
  :FIN        : Función a llamar una vez finalizada la partida, puede existir
                para cualquier tipo de agente. Recibe una lista con los agentes
                que han resultado muertos
  :ES-COMIDO  : Función para generador de plantas a llamar cuando una de sus
                plantas es comida
  :ES-MATADO  : Función para generador de direcciones a llamar cuando el agente
                es matado (pendiente de implementar)
  :SEG-POS    : Función de seguimiento de cambios de posición de agentes
                generadores de direcciones. Cada vez que un agente  generador
                de dirección cambia de posición es llamada. Puede existir para
                cualquier tipo de agente. Recibe el identificador del agente
                que se ha modificado y la nueva posición del agente
  :SEG-EST    : Función de seguimiento de cambios de estado de agentes
                generadores de direcciones. Cada vez que un agente cambia de
                dirección es llamada. Recibe el identificador del agente que se
                ha modificado y la nueva dirección del agente. Puede existir
                para cualquier tipo de agente
  :SEG-NPLA   : Función de seguimiento de creaciones de nueva planta. Cada vez
                que un agente generador de plantas genera una nueva planta es
                llamada. Puede existir para cualquier tipo de agente. Recibe el
                identificador del agente que ha creado la planta y la posición
                de la planta generada
  :SEG-MPLA   : Función de seguimiento de muerte de plantas. Cada vez que una
                planta muere (por ser comida o que la mata el propio agente
                generador de plantas) es llamada. Puede existir para cualquier
                tipo de agente. Recibe el identificador del agente que generó
                la planta muerta, el identificador del agente que mató la
                planta y la posición de ésta antes de que muriera
  :SEG-CPOSJ  : Función de seguimiento de consultas a posición de agentes. Cada
                vez que un agente realiza una consulta ENVIO:CONSULTA-POS-JUG
                es llamada. Recibe el identificador del agente que realiza la
                consulta y el identificador del agente sobre el que es
                realizada la consulta. Puede existir para cualquier tipo de
                agente
  :SEG-CESTJ  : Función de seguimiento de consultas a estado de agentes. Cada
                vez que un agente realiza una consulta ENVIO:CONSULTA-EST-JUG
                es llamada. Recibe el identificador del agente que realiza la
                consulta y el identificador del agente sobre el que es
                realizada la consulta. Puede existir para cualquier tipo de
                agente
  :SEG-CPOST  : Función de seguimiento de consultas a posiciones del tablero.
                Cada vez que un agente realiza una consulta
                ENVIO:CONSULTA-POS-TAB es llamada. Recibe el identificador del
                agente que realiza la consulta, la posición sobre la que se
                consulta y dirección. Puede existir para cualquier tipo de
                agente
  :REC-MSJ    : Función receptora de mensajes. Cada vez que un agente realiza
                una llamada a ENVIO:MENSAJE es llamada. Recibe el identificador
                del agente que realiza la llamada y el mensaje enviado .Puede
                existir para cualquier tipo de agente
  :SERVICIO   : Función que ofrece algún servicio no definido por parte del
                agente a otros agentes. Se trata de un servicio no identificado,
                esto es, serán llamadas de forma broadcast (ver ENVIO:BROADCAST)
                todas las funciones con igual nombre de servicio y no se espera
                que devuelvan un resultado al agente llamador. Los parámetros a
                recibir dependerán del servicio en cuestión. Puede existir para
                cualquier tipo de agente y será llamado desde cualquier tipo de
                agente
  :SERVICIO-ID: Función que ofrece algún servicio no definido por parte del
                agente a otros agentes. Se trata de un servicio identificado,
                esto es, serán llamadas de forma puntual (ver ENVIO:PUNTUAL)
                un agente con funcion con igual nombre de servicio y se espera
                que devuelva un resultado al agente llamador. Los parámetros a
                recibir dependerán del servicio en cuestión. Puede existir para
                cualquier tipo de agente y será llamado desde cualquier tipo de
                agente
") 

(defconstant keywords-fun-id-ag
  '(:inicio :turno :fin :es-comido :servicio-id)
  "Lista de keywords de tipos de funciones que en su llamada deben enviar el identificador del agente que es llamado (funciones identificadas).")

(defconstant keywords-nofun-ag
  '(:name :documentation :tipo-ag :use)
  "Lista de keywords que no son de tipos de funciones de agentes.")

;; Estructura de representación del problema a resolver
;; ----------------------------------------------------

(defstruct pt
  " - Estructura: PT (DEFSTRUCT MAT POS-JUG DIR-JUG)
     Package: PT-BASE

Estructura que almacena un laberinto PlanTron, junto a las posiciones de los jugadores y las direcciones en que se mueven. Consta de tres campos: MAT es una matriz de dos dimensiones, POS-JUG, DIR-JUG y ENE-JUG son listas.

La matriz MAT representa el laberinto y los bloque a mover, y tiene por lo tanto las dimensiones máximas de éste. Las paredes del laberinto (#) se representa con el símbolo contenido en la variable OBS (ver (help 'OBS). Los espacios vacios estan representados mediante NIL.

La mayoría de funciones de este paquete hace uso de este tipo de estructura."
  mat
  pos-jug
  dir-jug
  ene-jug)

(defvar *tablero* nil
  "Guarda le estructura tipo PT del juego actual")


;; Variables globales para tipo de partida
;; ---------------------------------------

(defvar num-jug-gdir nil
  "Número de jugadores generadores de direccion en la partida actual.")

(defvar num-jug-act nil
  "Número de jugadores activos (generadores de dirección + generadores de plantas) en la partida actual.")

(defvar pla-car nil
  "Indica si los jugadores generadores de plantas crean plantas carnivoras que matan a los generadores de direccion (T) o no (NIL)")

(defvar mov-con t
  "Indica si los jugadores generadores de dirección tienen movimiento contínuo (en última dirección elegida) (T) o si se puede detener (NIL)")

(defvar dim-tab nil
  "Dimensiones del la matriz MAT de la estructura PT (tablero).")

(defvar sin-bordes nil
  "Indica si el tablero tiene obstáculos en los bordes (NIL) o no (T)")

(defvar no-verbose nil
  "Indica si la partida actual han de informarse los cambios de posición o no.")

(defvar asincrono nil
  "Indica si la partida actual es asíncrona (NIL) o síncrona (T).")

(defvar sleep-cs nil
  "Indica las centesimas de segundo a esperar entre cada movimiento de jugadores.")

(defvar sleep-nil nil
  "Indica si desactivar el contador de SLEEP-CS centésimas de segundo entre movimientos (movimientos extra-rápidos).")


(defvar *num-jug-gdir* 1
  "Indica el número de jugadores generadores de dirección por defecto.")

(defvar *num-jug-act* (+ *num-jug-gdir* 1)
  "Indica el número de jugadores activos por defecto.")

(defvar *pla-car* nil
  "Indica si los jugadores generadores de plantas por defecto crean plantas carnivoras que matan a los generadores de direccion (T) o no (NIL)")

(defvar *mov-con* t
  "Indica si los jugadores generadores de dirección por defecto tienen movimiento contínuo (en última dirección elegida) (T) o si se puede detener (NIL)")

(defvar *dim-tab* '(80 23)
  "Indica las dimensiones del tablero de juego por defecto.")

(defvar *sin-bordes* nil
  "Indica si el tablero tiene obstáculos en los bordes (NIL) o no (T) por defecto.")

(defvar *no-verbose* nil
  "Indica si la partida han de informarse los cambios de posición o no por defecto.")

(defvar *asincrono* nil
  "Indica si la partida es asíncrona (NIL) o síncrona (T) por defecto.")

(defvar *sleep-cs* 15
  "Indica las centesimas de segundo a esperar entre cada movimiento de jugadores por defecto.")

(defvar *sleep-nil* nil
  "Indica si desactivar el contador de SLEEP-CS centésimas de segundo entre movimientos (movimientos extra-rápidos).")


;;
;; Otras variables globales
;; ------------------------

(defvar *ratio-ene-jug-ini* 0.025
  "Ratio de energia inicial de los jugadores generadores de dirección sobre el número de casillas del tablero.")

(defvar *ene-jug-gpla-ini* 0
  "Energía inicial de los jugadores generadores de plantas.")

(defvar end nil ;(1+ num-jug-act)
  "Contiene el número que representa la posición de los choques (!) dentro de la matiz del campo MAT de la estructura tipo PT.")


;; Funciones para el acceso a la estructura pt
;; -------------------------------------------

(defmacro pos-tab (tab listxy)
  "Posición x y del tablero"
  `(aref (pt-mat ,tab) (first ,listxy) (second ,listxy)))

(defmacro pos-jug (tab jug)
  "Devuelve la posición (x y) del jugador <jug> dentro del tablero <tab>"
  `(nth (1- ,jug) (pt-pos-jug ,tab)))

(defmacro dir-jug (tab jug)
  "Dirección de movimiento del jugador jug (^pos [0 1 ó -1] dir)"
  `(nth (1- ,jug) (pt-dir-jug ,tab)))

(defmacro ene-jug (tab jug)
  "Devuelve la energía del jugador <jug>"
  `(nth (1- ,jug) (pt-ene-jug ,tab)))

(defmacro pos-inc-jug (tab jug)
  "Dirección a incrementar del jugador jug (^pos ó nil)"
  `(first (first (dir-jug ,tab ,jug))))

(defmacro val-inc-jug (tab jug)
  "Suma a incrementar del jugador jug [0, 1 ó -1]"
  `(second (dir-jug ,tab ,jug)))

(defmacro dir-dir-jug (tab jug)
  "Dirección numérica del jugador jug [nil, 0, 1, 2 ó 3]"
  `(third (dir-jug ,tab ,jug)))

(defun borde-out-tab (posxy)
  "SI posición POSXY está en un borde del tablero o fuera de él devuelve T, sino NIL."
  (or (>= 0 (first posxy))
      (>= 0 (second posxy))
      (<= (1- (first dim-tab)) (first posxy))
      (<= (1- (second dim-tab)) (second posxy))))


;; Generador de plantas
;; --------------------

(defun act-planta (mov)
  "Coloca (FIRST MOV) (JUG) en la posición (SECOND MOV) POSXY del tablero si es posible incrementando la energia del jugador en (THIRD MOV) INC-ENE, donde JUG ha de ser una planta. Devuelve MOV si se ha podido hacer o NIL en caso contrario"
  (let ((jug (first mov))
	(posxy (second mov))
	(inc-ene (third mov)))
    (if (and (> jug num-jug-gdir)
	     (consp posxy)
	     (numberp (first posxy))
	     (numberp (first posxy))
	     (not (borde-out-tab posxy)))
	(progn
	  (setf (pos-tab *tablero* posxy) jug)
	  (if (numberp inc-ene)
              ; Incrementar energía generador
	      (setf (ene-jug *tablero* jug) (+ (ene-jug *tablero* jug) inc-ene)))
	  mov))))


(defun indicar-colision (jug-gpla jug-gdir posxy)
  "Indicar al JUG-GPLA el choque de JUG-GDIR en posición POSXY"
  (envio:puntual jug-gpla :es-comido jug-gdir posxy)
  (envio:broadcast :seg-mpla jug-gpla jug-gdir posxy))


;; Obtención de movimientos
;; ------------------------

(defvar lis-mov nil    ; No usado
  "Lista de movimientos a efectuar tanto de gendir como genpla")

(defvar lis-mov-gdir nil
  "Lista de movimientos a efectuar de gendir")

(defvar lis-mov-gpla nil
  "Lista de movimientos a efectuar de genpla")

(defvar lis-jug nil    ; No útil
  "Lista de jugadores (total de agentes activos)")

(defvar lis-jug-gdir nil
  "Lista de jugadores tipo gdir")

(defvar lis-jug-gpla nil
  "Lista de jugadores tipo gpla")

(defun get-accion (jug)
  "Lanza a jugador con identificador JUG una solicitud de acción, almacenado su resultado para un posterior proceso"
  (if asincrono
      (let ((accion (envio:puntual jug :turno)))
	(if accion
	    (progn
	      (push (list jug accion) lis-mov-gdir)
	      t)))))

(defun gen-planta (jug posxy &optional inc-ene)
  "Petición de colocar JUG en la posición POSXY del tablero, donde JUG se trata de una planta, para un posterior proceso"
  (push (act-planta (list jug posxy inc-ene)) lis-mov-gpla)
  t)

(defun obt-lis-mov-gdir ()
  "Obtiene lista de movimientos de los jugadores gdir"
  (if asincrono
      (prog1
	  lis-mov-gdir
	  (setf lis-mov-gdir nil))
    (prog1
	(mapcar #'(lambda (jug)
		    (list jug
			  (envio:puntual jug :turno)))
		lis-jug-gdir)
        (setf lis-mov-gdir nil))))

(defun obt-lis-mov-gpla ()
  "Obtiene lista de movimientos de los jugadores gpla"
  (if asincrono
      (prog1
	  lis-mov-gpla
	  (setf lis-mov-gpla nil))
    (prog1
	(nconc
	 (mapcar #'(lambda (jug)
			   (list jug
				 (envio:puntual jug :turno)))
		       lis-jug-gpla)
	 lis-mov-gpla)
        (setf lis-mov-gpla nil))))

(defun obt-lis-mov ()  ; No usado
  "Obtiene lista de movimientos de los jugadores (gdir y gpla)"
  (if asincrono
      (prog1
	  lis-mov
	  (setf lis-mov nil))
    (prog1
	(nconc ; Llamamos los turnos, primero gplas y luego gdirs
	       (mapcar #'(lambda (jug)
			   (list jug
				 (envio:puntual jug :turno)))
		       lis-jug-gpla)
	       lis-mov
	       (mapcar #'(lambda (jug)
			   (list jug
				 (envio:puntual jug :turno)))
		       lis-jug-gdir))
        (setf lis-mov nil))))


;; Instrucciones del juego
;; -----------------------

(defconstant dir-inc
  (make-array '(4 2) :initial-contents
	      '((1 -1)    ; Norte
		(0 +1)    ; Este
		(1 +1)    ; Sur
		(0 -1)))  ; Oeste
  "Array para transformar de dirección a campo dir-jug de pt")

(defmacro dir2dim-inc (dir)
  `(aref dir-inc ,dir 0))

(defmacro dir2pos-inc (tab jug dir)
  `(nthcdr (dir2dim-inc ,dir) (pos-jug ,tab ,jug)))

(defmacro dir2val-inc (dir)
  `(aref dir-inc ,dir 1))

(defmacro jug2dim-inc (tab jug)
  `(nth (aref dir-inc (dir-dir-jug ,tab ,jug) 0) dim-tab))

(defun change-dir (tab jug dir)
  "Cambia la dirección de un jugador gendir y devuelve algo verdadero (la nueva dirección DIR o T si la dirección cambiada es NIL) o NIL si la dirección era la misma. Si el jugador no es tipo gendir devuelve NIL"
  (if (<= jug num-jug-gdir)
      (let ((dirj (dir-jug tab jug)))
	(if (not (eq (third dirj) dir))
	    (if (and (not mov-con) (not dir))
		(progn
		  (setf (first dirj) nil)
		  (setf (second  dirj) 0)
		  (setf (third dirj) dir)
		  t)
	      (if (and (member dir '(0 1 2 3)))
		  (progn
		    (setf (first dirj) (dir2pos-inc tab jug dir))
		    (setf (second  dirj) (dir2val-inc dir))
		    (setf (third dirj) dir))))))))

(defun test-mark-mat-pos (tab posxy mark-nil &optional mark-no-nil)
    "Marcamos en la posición posxy de matriz, si está a nil, con mark, si no con mark-no-nil y devolvemos nil o t respectivamente"
    (if (pos-tab tab posxy)
	(progn
	  (if mark-no-nil
	      (setf (pos-tab tab posxy) mark-no-nil))
	  t)
      (progn (setf (pos-tab tab posxy) mark-nil)
	     nil)))

(defun manten-dentro (val lon-dim)
  "Si se sale del tablero volver por otro lado."
  (cond
   ((>= val lon-dim) (rem val lon-dim))
   ((< val 0) (rem (+ lon-dim val) lon-dim))
   (t val)))

(defun check-set-inc-pos (tab)
  "Incrementa las posiciones de los jugadores gen-dir y chequea si hay conflicto (colisión), devuelve nil o lista con jugadores en conflicto"
  ; Avanzamos los jugadores consumiendo energía
  (dotimes (i num-jug-gdir)
    (let ((jug (1+ i)))
      (if (or mov-con (dir-dir-jug tab jug))
	  (progn
	    (setf (pos-inc-jug tab jug)
		  (manten-dentro (+ (pos-inc-jug tab jug) (val-inc-jug tab jug))
				 (jug2dim-inc tab jug)))
	    (setf (ene-jug tab jug) (1- (ene-jug tab jug)))))))
  ; Chequeamos conflictos
  (do* (res
	(jug 1 (1+ jug))
	(list-pos-jug-ant nil (cons (first list-pos-jug-pos) list-pos-jug-ant))
	(list-pos-jug-pos (pt-pos-jug tab) (rest list-pos-jug-pos))
	(posxy (first list-pos-jug-pos) (first list-pos-jug-pos)))
      ((endp list-pos-jug-pos) (reverse res))
    (if (or (member posxy (rest list-pos-jug-pos) :test #'equal)
	    (member posxy list-pos-jug-ant :test #'equal))
	; Choca con otro gdir, marcar con end y añade a lista de chocados
	(progn
	  (test-mark-mat-pos tab posxy end end)
	  (pushnew jug res))
      ; No choca con otro gdir
      (progn
	    ; Mira si choca rastro gdir y gpla y marca
	(if (test-mark-mat-pos tab posxy jug)
                ; Choca con obs o jug-gdir
	    (if (<= (pos-tab tab posxy) num-jug-gdir)
                    ; No choca contra si mismo sin moverse
		(if (or mov-con (not (eq jug (pos-tab tab posxy))) (dir-dir-jug tab jug))
		    (progn (test-mark-mat-pos tab posxy jug end)
			   (pushnew jug res)))
	      ; Choca con jug-gpla
	      (if pla-car
		  ; Plantas carnivoras, matar gdir
		  (progn
		    (test-mark-mat-pos tab posxy end end)
		    (pushnew jug res))
	        ; Plantas no carnivoras, indicar a jug-gpla y marcar
		(progn
		  (indicar-colision (pos-tab tab posxy) jug posxy)
		  (test-mark-mat-pos tab posxy jug jug)))))
        ; Mira si energía a cero y mata
	(if (and (zerop (ene-jug tab jug)))
	    (progn (test-mark-mat-pos tab posxy jug end)
		   (pushnew jug res)))))
      ; Modo asíncro y sin movimiento continuo, parar de mover
      (if (and asincrono (not mov-con))
	  (change-dir tab jug nil))))


;; Problema inicial e interno
;; --------------------------

(defun jug1-pos-ini ()
    "Posición inicial del jugador 1"
  (list (floor (1- (first dim-tab)) 4) (floor (1- (second dim-tab)) 2)))

(defun jug2-pos-ini ()
    "Posición inicial del jugador 2"
  (list (ceiling (* (1- (first dim-tab)) 3) 4) (floor (1- (second dim-tab)) 2)))

(defun jugs-pos-ini (n)
  "Posiciones iniciales de los N jugadores"
  ; Ojo, solo admite 2 variantes (8 jugadores)
  (let (pos-ini)
    (dotimes (i (ceiling n 2) (reverse (nthcdr (rem n 2) pos-ini)))
      (push (jug1-pos-ini) pos-ini)
      (push (jug2-pos-ini) pos-ini))))

(defconstant jug1-dir-ini 1   ; Este
  "Dirección inicial del jugador 1 en el tablero por defecto")

(defconstant jug2-dir-ini 3   ; Oeste
  "Dirección inicial del jugador 2 en el tablero por defecto")

(defun jugs-dir-ini (n)
  "Direcciones iniciales de los N jugadores"
  ; Ojo, solo admite 4 variantes (8 jugadores)
 (let (dir-ini
       (cont1 jug1-dir-ini)
       (cont2 jug2-dir-ini))
    (dotimes (i (ceiling n 2) (reverse dir-ini))
      (push cont1 dir-ini)
      (push cont2 dir-ini)
      (setf cont1 (rem (1- cont1) 4))
      (setf cont2 (rem (1+ cont2) 4)))))

(defun mark-borde-tab (tab dim-tab)
  "Marcar los bordes del tablero"
  (dotimes (x (first dim-tab))
    (test-mark-mat-pos tab (list x 0) obs)
    (test-mark-mat-pos tab (list x (1- (second dim-tab))) obs))
  (dotimes (y (second dim-tab))
    (test-mark-mat-pos tab (list 0 y) obs)
    (test-mark-mat-pos tab (list (1- (first dim-tab)) y) obs)))


;; Generacion estructura inicial del problema a resolver
;; -----------------------------------------------------

(defun gen-pt-def (num-jug-gdir num-jug-act dim-tab sin-bordes mov-con)
  "Genera tablero por defecto de jugadores y dimensiones pasados"
  (let ((tab-ini (make-pt)))
    ; Inicializamos los campos de la estructura tab-ini
    (setf (pt-mat tab-ini) (make-array dim-tab :initial-element nil)
	  (pt-pos-jug tab-ini) (jugs-pos-ini num-jug-gdir)
	  (pt-dir-jug tab-ini) nil
	  (pt-ene-jug tab-ini) nil)
    ; Colocamos direccion de movimiento inicial a nil
    (dotimes (i num-jug-gdir)
      (push (list nil 0 nil) (pt-dir-jug tab-ini)))
    (if mov-con ; Con movimiento continuo dirección inicial a ...
	(let ((dirs (jugs-dir-ini num-jug-gdir)))
	  (dotimes (i num-jug-gdir)
	    (change-dir tab-ini (1+ i) (nth i dirs)))))
    ; Colocamos la energia inicial de los jugadores
    (let ((lis-ene nil)
	  (ene-ini-gdir (round (* *ratio-ene-jug-ini* 
				  (first dim-tab)
				  (second dim-tab)))))
      (dotimes (i num-jug-gdir)
	(push ene-ini-gdir lis-ene))
      (dotimes (i (- num-jug-act num-jug-gdir))
	(push *ene-jug-gpla-ini* lis-ene))
      (setf (pt-ene-jug tab-ini) (reverse lis-ene)))
    ; Marcamos en la matriz las posiciones iniciales de jugadores
    (dotimes (i num-jug-gdir)
      (test-mark-mat-pos tab-ini (pos-jug tab-ini (1+ i)) (1+ i)))
    (if (not sin-bordes)
        ; Marcamos en la matriz los bordes del tablero
	(mark-borde-tab tab-ini dim-tab))
    ; Devolvemos tablero
    tab-ini))

 
(defun gen-pt (num-jug-gdir num-jug-act dim-tab sin-bordes mov-con)
  "Genera un estado inicial con un laberinto para el PlanTron, de acuerdo a las especificiaciones del argumento TIPO pasado a la función INIC-PT, puede cargar el laberinto desde un fichero si asi se le ha especificado en esta variable (ver (help 'INIC-PT). Devuelve una estructura tipo PT."
  (gen-pt-def num-jug-gdir num-jug-act dim-tab sin-bordes mov-con))


;;
;; Interface de consultas a servidor de juego
;; ------------------------------------------

(defun leyenda ()
  "Consulta la leyenda que define el juego actual implementado."
  (format nil "~A-~A" *juego* version-entorno))

(defun pos-jugador (jugador)
  "Consulta posición actual del jugador con identificador JUGADOR. El resultado, si JUGADOR corresponde al identificador de un jugador generador de direcciones válido, es una lista con dos números correspondientes al valor en las coordenadas X e Y. Las coordenadas comienzan (0 0) en la esquina superior izquierda. Si JUGADOR es un identificador de un jugador no válido se devuelve NIL."
  (copy-list (pos-jug *tablero* jugador)))

(defun dir-jugador (jugador)
  "Consulta dirección actual del jugador con identificador JUGADOR. El resultado, si JUGADOR corresponde al identificador de un jugador válido, es un número entre 0 y 4, correspondientes a las direcciones de Norte, Este, Sur y Oeste respectivamente. Si JUGADOR es un identificador de un jugador no válido se devuelve NIL."
  (dir-dir-jug *tablero* jugador))

(defun pos-inc-dir (pos dir)
  "Calcula la posición resultante de desplazarse en la direcciónb DIR respecto a POS."
  (let ((pos-inc (copy-list pos)))
    (setf (nth (dir2dim-inc dir) pos-inc)
	  (+ (nth (dir2dim-inc dir) pos-inc)
	     (dir2val-inc dir)))
    pos-inc))

(defun ene-jugador (jugador) ; FIXME ya esta en est-jug ¿eliminar?
  "Consulta la energía actual del jugador con identificador JUGADOR. El resultado, si JUGADOR corresponde al identificador de un jugador válido, es un número entero mayor o igual a 0. Existe una energía inicial igual para cada jugador, se consume una unidad de enrgía por cada paso que da el jugador y cuando la engergía llega a 0 el jugador muere. Si JUGADOR es un identificador de un jugador no válido se devuelve NIL."
  (ene-jug *tablero* jugador))

(defun est-jugador (jugador)
  "Consulta estado actual del jugador con identificador JUGADOR. El resultado, si JUGADOR corresponde al identificador de un jugador válido, es una lista con la dirección del jugador (calculada con DIR-JUGADOR) y la energia del jugador (calculada con ENE-JUGADOR)."
  (list (dir-jugador jugador) (ene-jugador jugador)))

(defun transf-ene (jugador-org jugador-des cantidad)
  "Realiza una transferencia de energía de JUGADOR-ORG a JUGADOR-DES de CANTIDAD unidades de energía, si es posible. Si CANTIDAD es no es un número positivo transfiere toda la energía del JUGADOR-ORG. Devolviendo la cantidad de energía transferida si se pudo o NIL si falla."
  (if (and (numberp (ene-jugador jugador-org)) (numberp (ene-jugador jugador-des)))
      (if (and (numberp cantidad) (or (>= cantidad 0)))
	  (let ((cantidad-real (min (ene-jugador jugador-org) cantidad)))
	    (setf (ene-jug *tablero* jugador-org) (- (ene-jug *tablero* jugador-org)
						     cantidad-real)
		  (ene-jug *tablero* jugador-des) (+ (ene-jug *tablero* jugador-des)
						     cantidad-real))))))

(defun pos-tablero (pos &optional dir)
  "Contenido actual de una posición del tablero POS más opcionalmente desplazamiento en una dirección DIR sobre ella. Los valores de POS y DIR han de ser como se indican en POS-JUGADOR y en DIR-JUGADOR. El resultado devuelto es NIL si la casilla está vacía. Si la casilla está ocupada se devuelve el identificador del jugador que marcó dicha casilla, o el contenido de OBS en caso de que contenga un obstáculo (borde del tablero)."
  ; Chequeo de parámetros
  (if (and (consp pos) (numberp (first pos)) (numberp (second pos))
	   (or (null dir) 
	       (and (numberp dir) (>= dir 0) (<= dir 3))))
      ; Cálculo de posición a chekear
      (let ((pos-to-check 
	     (if (not dir)
		 pos
	       (pos-inc-dir pos dir))))
        ; Chequeo de bordes y resultado
	(if (borde-out-tab pos-to-check)
	    (if sin-bordes
		(progn
		  (setf (first pos-to-check)
			(manten-dentro (first pos-to-check)
				       (first dim-tab))
			(second pos-to-check)
			(manten-dentro (second pos-to-check)
				       (second dim-tab)))
		  (pos-tab *tablero* pos-to-check))
		obs)
	  (pos-tab *tablero* pos-to-check)))))

(defun tipo-jug (id)
  "Tipo actual de ID, que puede ser el contenido del tablero, incluyendo por tanto un identificador de jugador, NIL (libre) o OBS (obstaculo). El tipo devuelto puede ser uno de los siguientes símbolos:
 nil: Posición libre
 obstaculo: obstáculo fijo del tablero
 choque: indica que hay una colision de varios jugadores
 jugador: jugador o su rastro
 planta: planta
 desconocido: idem"
  (cond
   ((null id)                 nil)
   ((numberp id)
    (cond
     ((= id obs)             'obstaculo)
     ((= id end)             'choque)
     ((<= id num-jug-gdir)   'jugador)
     (t                      'planta)))
   (t                        'desconocido)))

(defun copy-array (arr)
  "Genera una copia de un array de dos dimensiones."
  (let ((arr2 (make-array `(,(array-dimension arr 0)
		,(array-dimension arr 1)))))
    (dotimes (i (array-dimension arr 0) arr2)
      (dotimes (j (array-dimension arr 1))
	(setf (aref arr2 i j) (aref arr i j))))))

(defun tablero ()
  "Solicita la matriz del tablero PlanTron, devolviéndolo en un array."
  (copy-array (pt-mat *tablero*)))


;; Contador de tiempo entre movimientos
;; ------------------------------------

(defvar counter nil
  "Vairable contador.")

(defun ini-counter (limit)
  "Inicializa el contador a LIMIT centésimas de segundo. Si LIMIT es NIL el contador queda nulo."
  (setf counter
	(if limit
	    (+ (get-internal-real-time)
	       (/ (* limit internal-time-units-per-second)
		  100)))))

(defun test-counter (&optional hang)
  "Chequea si el contador llegado al límite, si es así devuelve T, sino devuelve NIL a menos que HANG sea T, que espera hasta que se llegue al límite. Si el contador está nulo devuelve HANG."
  (if counter
      (do ()
	  ((> (get-internal-real-time) counter) t)
	(if (not hang)
	    (return nil)))
    hang))


;; Lanzador de agentes y problemas
;; -------------------------------

(defun gen-estad-ag (tiempo-final)
  "Genera las estadísticas de resultados de los agentes"
  (let ((estad nil))
    (dotimes (i num-jug-act (reverse estad))
      (let ((jug (1+ i)))
	(push (list jug tiempo-final (ene-jug *tablero* jug)) estad)))))

(defmacro getf-set (plist key val-def)
  "Extrae el valor de la clave KEY en la lista PLIST, en el caso de no estar inserta dicha clave con el valor VAL-DEF en PLIST y devuelve VAL-DEF."
  `(if (eq (getf ,plist ,key 'noesta) 'noesta)
      (setf (getf ,plist ,key) ,val-def)
    (getf ,plist ,key)))

(defun extrc-tipo-partida (tipo-partida)
  "Extracción del tipo de partida hacia variables globales y relleno de los campos por defecto. TIPO-PARTIDA puedes ser una lista de las siguientes keywords:
  :NUM-JUG-GDIR : Número de jugadores generadores de direcciones
  :NUM-JUG-ACT  : Número de jugadores activos (generadores direcciones y de plantas)
  :PLA-CAR      : Plantas carnivoras o no
  :MOV-CON      : Movimiento continuo de gendirs
  :DIM-TAB      : Lista con dimensiones del tablero (X Y)
  :SIN-BORDES   : Si no se colocan bordes alrededor del tablero
  :SLEEP-CS     : Centésimas de segundo a transcurrir entre cada turno
                  de juego
  :NO-VERBOSE   : Si no emite seguimiento de cambio de posiciones
  :ASINCRONO    : Si juego es asíncrono (no implementado)
  :SLEEP-NIL    : Si desactivación del contador de tiempo entre turnos
"
  (setf num-jug-gdir (getf-set tipo-partida :num-jug-gdir *num-jug-gdir*))
  (setf num-jug-act  (getf-set tipo-partida :num-jug-act *num-jug-act*))
  (setf pla-car      (getf-set tipo-partida :pla-car *pla-car*))
  (setf mov-con      (getf-set tipo-partida :mov-con *mov-con*))
  (setf dim-tab      (getf-set tipo-partida :dim-tab *dim-tab*))
  (setf sin-bordes   (getf-set tipo-partida :sin-bordes *sin-bordes*))
  (setf sleep-cs     (getf-set tipo-partida :sleep-cs *sleep-cs*))
  (setf no-verbose   (getf-set tipo-partida :no-verbose *no-verbose*))
  (setf asincrono    (getf-set tipo-partida :asincrono *asincrono*))
  (setf sleep-nil    (getf-set tipo-partida :sleep-nil *sleep-nil*))
  tipo-partida)

(defun inicio (&optional tipo-partida)
  "Lanza una partida PlanTron del tipo TIPO-PARTIDA (ver EXTRC-TIPO-PARTIDA)."
  ; Establecemos var. globales y valores por defecto del tipo de juego
  (setf tipo-partida (extrc-tipo-partida tipo-partida))
  ; Generamos tablero de juego
  (setf *tablero* (gen-pt num-jug-gdir num-jug-act dim-tab sin-bordes mov-con))
  ; Establecemos valor para marca de final
  (setf end (1+ num-jug-act))
  ; Inicializamos contador de espera entre turnos
  (ini-counter (and (not sleep-nil) sleep-cs))
  ; Inicializamos lista de jugadores activos y gendir
  (setf lis-jug nil)
  (setf lis-jug-gdir nil)
  (setf lis-jug-gpla nil)
  (dotimes (i num-jug-gdir)
    (push (1+ i) lis-jug)
    (push (1+ i) lis-jug-gdir))
  (dotimes (i (- num-jug-act num-jug-gdir))
    (push (+ i 1 num-jug-gdir) lis-jug-gpla)
    (push (+ i 1 num-jug-gdir) lis-jug))
  ; Inicializamos lista de movimientos
  (setf lis-mov-gdir nil)
  (setf lis-mov-gpla nil)
  ; Inicializamos los agentes
  (envio:broadcast :inicio tipo-partida)

  ; Inicio bucle iteraciones --------------------------------------------
  (do* ((tiempo 0 (1+ tiempo))
	lis-movs-gpla
	lis-movs-gdir
                     ; Incrementa posiciones y chequea fin
	(muertos nil (check-set-inc-pos *tablero*)))
       ; Condición de salida ----------------------------------------------
      (muertos
       ; Indicamos a los agentes fin de partida,
       ; ojo, FIXME, los últimos movimientos no se informan
       (envio:broadcast :fin muertos)
       (list muertos 
	     (gen-estad-ag tiempo)))
    ; Cuerpo del bucle --------------------------------------------------
    ; Turnos y obtención de movimientos gpla
    (setf lis-movs-gpla (remove-if-not #'(lambda (mov)
					   (if (consp (second mov))
					       mov))
				       (obt-lis-mov-gpla)))
    ; Informa de cambios de posición jugadores genpla
    (if (not no-verbose)
	(mapcar #'(lambda (cpos)
		    (envio:broadcast
		     :seg-npla
		     (first cpos)
		     (second cpos)))
		lis-movs-gpla))
    ; Turnos y obtención de movimientos gdir
    (setf lis-movs-gdir (remove-if-not #'(lambda (mov)
					   (change-dir *tablero*
						       (first mov)
						       (second mov)))
				       (obt-lis-mov-gdir)))
    ; Informa de cambios de dirección
    (mapcar #'(lambda (mov)
		(envio:broadcast
		 :seg-est
		 (first mov)
		 (second mov)))
	    lis-movs-gdir)
    ; Informa de cambios de posición jugadores gendir
    (if (not no-verbose)
	(mapcar #'(lambda (jug)
		    (envio:broadcast
		     :seg-pos
		     jug
		     (pos-jug *tablero* jug)))
		lis-jug-gdir))
    ; Esperamos por el contador
    (test-counter t)
    (ini-counter (and (not sleep-nil) sleep-cs))))
    ; Fin del bucle de iteraciones --------------------------------------
