;; Programa: Package con agentes pasivos con distintas
;;           utilidades para asistir a agentes jugadores de
;;           PlanTron
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    18/05/2007
;; Comentarios:
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT, PT-BASE Y ENVIO.
;;
;;           Las funciones accesibles más importantes son:
;;
;;           ....
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;;           Hace uso de los paquetes para crear sistemas de
;;           prudocción (SHELL) y búsquedas genéricas
;;           (GENERIC-SEARCH)
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (18/05/2007): Primera versión.
;;
;;           - Ver 0.1.1 (25/05/2007): Añadido keys :servicio-id
;;             Reparado error en PLANTA-CERCA que a veces devolvía
;;             menos plantas de las pedidas e inicializaba mal
;;
;;           - Ver 0.2 (28/05/2008): Mejorado eficiencia de
;;             PLANTA-CERCA según variante
;;             [Fukunaga and Narendra, 1975]
;;
;;           - Ver 0.2.1 (01/06/2008): Incorporada cache para 
;;             acelerar las búsquedas a PLANTA-CERCA
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - Funciones #:GENERIC-SEARCH, #:REGLA, #:LIMPIA-REGLAS
;;             e #:HIJOS no cogen bien el nombre del pacakge en la
;;             definición de AGENTE
;;           - Eliminar warnings de compilación
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Incorporar gestión de energía de muertes y creaciones
;;             de plantas
;;           - Generalizar planta cerca para que pueda encontrar no
;;             solo plantas, sino incluso otros objetos
;;           - Que se pueda indicar de alguna manera los objetos a
;;             seguir (lista de id's)
;;           - Identificador de contador y/ contador por id agente
;;           - Mejorar interfaces con shell y generic-search para
;;             permitir que varios agentes las usen en la misma
;;             partida
;;           - Cambiar todos los :servicio-id por key concreta y
;;             ver si implementar en llamadas
;;           - Generar agente que reciba e imprima :mensaje en un
;;             dispositivos configurable ("/dev/pts/?")
;;           - ¿Que planta cerca solo guarde en su lista las plantas
;;             que no estén rodeadas de otras? Así es más fácil
;;             encontrar la planta más cercana
;;           - Relizar caché de llamadas a planta-cerca
;;           - Ajustar valor de rad-max en función de las dimensiones
;;             del tablero
;;
;;           - Mejoras en estilo de programación
;;             ---------------------------------
;;           - Plantear cambiar interface de PLANTA-CERCA para que
;;             las plantas devueltas sean listas con:
;;             (distancia posicion identificador)
;;           - Eliminar agentes planta-cerca3 y 2 una vez validado
;;             planta-cerca y *-old
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "PT-UTIL"
  (:nicknames "PT-UT")
  (:use "COMMON-LISP")
  (:export "AGENTES" "VERSION" "*USAR-CACHE-PC*" "*LONMAX-CACHE-PLA*" "PLANTA-CERCA" "PLANTAS-CERCA-INI" "PLANTA-CERCA-NUEVA-PLANTA" "PLANTA-CERCA-MUERE-PLANTA" "PLANTA-CERCA2" "PLANTAS-CERCA2-INI" "PLANTA-CERCA2-NUEVA-PLANTA" "PLANTA-CERCA2-MUERE-PLANTA" "PLANTA-CERCA3" "DISTANCIA-CITY" "CONSULTA-CONTADOR" "INICIA-CONTADOR" "CONTADOR-INI" "*SISTEMA*" "LIMPIA-REGLAS" "REGLA" "HIJOS" "*SEARCH-QUEUE*" "GENERIC-SEARCH" "*EVAL-FN*" "SPLICE" "PRIORITY-MERGE-FN" "PRIORITY-DEQUEUE-FN"))


(in-package "PT-UTIL")


;;
;; Constante versión
;; -----------------

(defconstant version "0.2.1"
  "Versión del package PT-UTIL.")


;;
;; Constante con lista de agentes presentes en este package
;; --------------------------------------------------------

(defconstant agentes
  `((:name "SHELL"
     :documentation ,(documentation (find-package "SHELL") T)
     :servicio-id (regla limpia-reglas hijos)
     :tipo-ag envio:pasivo)
    (:name "GENERIC-SEARCH"
     :documentation ,(documentation (find-package "GENERIC-SEARCH") T)
     :servicio-id generic-search
     :tipo-ag envio:pasivo)
    (:name "A_STAR"
     :documentation ,(documentation (find-package "A_STAR") T)
     :servicio-id (priority-a-merge-fn priority-a-dequeue-fn)
     :tipo-ag envio:pasivo)
    (:servicio-id (planta-cerca distancia-city)
     :name "PLANTA-CERCA"
     :use ("GENERIC-SEARCH" "A_STAR")
     :inicio plantas-cerca-ini
     :seg-npla planta-cerca-nueva-planta
     :seg-mpla planta-cerca-muere-planta
     :tipo-ag envio:pasivo)
    (:servicio-id (planta-cerca2 distancia-city)
     :name "PLANTA-CERCA2"
     :inicio plantas-cerca2-ini
     :seg-npla planta-cerca2-nueva-planta
     :seg-mpla planta-cerca2-muere-planta
     :tipo-ag envio:pasivo)
    (:servicio-id (planta-cerca distancia-city)
     :name "PLANTA-CERCA3"
     :use ("GENERIC-SEARCH" "A_STAR")
     :inicio plantas-cerca3-ini
     :seg-npla planta-cerca3-nueva-planta
     :seg-mpla planta-cerca3-muere-planta
     :tipo-ag envio:pasivo)
    (:name "CONTADOR"
     :documentation "Utilidad para cronometrar tiempos."
     :servicio-id (consulta-contador inicia-contador)
     :inicio contador-ini
     :tipo-ag envio:pasivo))
  "Agentes con utilidades para asistira a agentes jugadores de PlanTron.")



;; Agente de apoyo de sistemas de producción
;; -----------------------------------------

(defconstant lis-exp-shell '(shell:*sistema* shell:limpia-reglas shell:regla shell:hijos))
(shadowing-import lis-exp-shell)
(export lis-exp-shell)


(defconstant lis-exp-generic-search '(generic-search:*search-queue* generic-search:generic-search generic-search:*eval-fn* generic-search:splice priority-merge-fn generic-search:priority-dequeue-fn))
(shadowing-import lis-exp-generic-search)
(export lis-exp-generic-search)


(defconstant lis-exp-a_star '(a_star:*coste-acum-fn* a_star:*igual-est-fn* a_star:criba-iguales a_star:fusion a_star:priority-a-merge-fn a_star:priority-a-dequeue-fn))
(shadowing-import lis-exp-a_star)
(export lis-exp-a_star)


;;
;; Variables para agentes buscadores de plantas cercanas
;; -----------------------------------------------------

; Para buscador de plantas cercanas principal

(defstruct cluster
  "Esctructura para almacenar un cluster jerarquico de plantas"
  (set nil)      ; Lista de clusters o plantas asociadas (Sp)
  (num-pla 0)    ; Numero de plantas del cluster (Np)
  (pos nil)      ; Centroide (posición (x y)) del cluster (Mp)
  (rad 0)        ; Radio máximo del cluster (rp)
  (d-est nil))   ; Estimación de distancia (d(X,Mp) - rp)

(defvar arbol-plantas nil 
  "Nodo raiz del árbol de clusters jerárquicos")

(defvar *rad-max* 3
  "Radio máximo de los clusters mas simples")

(defvar dis-max nil
  "Distancia máxima alcanzable")

(defvar pos-jug nil
  "Posición a partir de que se buscará planta cerca")

(defvar id-pla nil
  "Identificador de las plantas a buscar, NIL si cualquiera")

(defvar num-pla 1
  "Número de plantas a buscar")

(defvar lis-pla-cerca nil
  "Lista de plantas cerca encontradas")

; Para buscador de plantas 2 cercanas

(defvar lis-plantas nil ; Cada planta (time-stamp (posx posy) id)
  "Lista de plantas actuales")

; Para caché de llamadas a planta-cerca

(defvar *usar-cache-pc* t
  "Indica si se usará la caché de llamadas a planta cerca")

(defvar ult-ts-lis-plantas 0
  "Valor último time-stamp de la lista de plantas")

(defvar *lonmax-cache-pla* 38
  "Número máximo de plantas a almacentar en lis-plantas")

(defvar lis-plantas-muertas nil ; Cada planta (time-stamp (posx posy) id)
  "Lista de plantas muertas")

(defvar time-stamp nil ; Se incrementa al cambiar time-cons-pc
  "Identificador temporal de turnos")

(defvar time-cons-pc nil
  "Indica si se está en estado de consultas de planta cerca o en estado de inserción y/o borrado de plantas, para incrementar el time-stamp")

(defvar mat-cache-pc nil
  ; Contiene (id num time-stamp <resultado ll planta-cerca>)
  "Matriz con caché de resultados de llamadas a planta cerca")


;;
;; Funciones de apoyo para caché de llamadas a planta-cerca
;; --------------------------------------------------------

; Funciones de plantas (pla)

(defmacro make-pla (pos id)
  `(list ,pos ,id))

(defmacro pla-pos (pla)
  `(first ,pla))

(defmacro pla-id (pla)
  `(second ,pla))

(defun eq-pla (pla1 pla2)
  "Indica si dos plantas son iguales"
  (and (equal (pla-pos pla1) (pla-pos pla2))
       (or (not (pla-id pla1))
           (not (pla-id pla2))
           (eq (pla-id pla1) (pla-id pla2)))))

; Funciones distancia e inserta ordenado

(defun distancia-city (pos1 pos2)
  "Calcula la distancia city block entre POS1 y POS2"
  (+ (abs (- (first pos1) (first pos2)))
     (abs (- (second pos1) (second pos2)))))

(defun inserta-ord (pla dist lis-ord num)
  "Inserta ordenada por distancia la planta en la lista LIS-ORD."
  (do* ((ant nil lis)
	(lis lis-ord (rest lis))
	(cont num (1- cont)))
      
      ((or (zerop cont)
	   (endp lis)
	   (<= dist (second (first lis))))
       (if (not (zerop cont))
	   (if ant
	       (setf (rest ant) (cons (list (pla-pos pla) dist (pla-id pla))
				      (rest ant)))
	     (setf lis-ord (cons (list (pla-pos pla) dist (pla-id pla))
				 lis-ord))))
       lis-ord)))


; Funciones sobre llamadas a planta cerca (llpc)

(defmacro make-llpc (id num ts res)
  `(list ,id ,num ,ts ,res))

(defmacro llpc-id (llpc)
  `(first ,llpc))

(defmacro llpc-num (llpc)
  `(second ,llpc))

(defmacro llpc-ts (llpc)
  `(third ,llpc))

(defmacro llpc-res (llpc)
  `(fourth ,llpc))

(defmacro llpc-ts-res (llpc)
  `(rest (rest ,llpc)))

(defun eq-llpc (llpc num id)
  (and llpc (= (length (llpc-res llpc)) num)  ; Encontrada llpc anterior
       (eq num (llpc-num llpc))
       (eq id (llpc-id llpc))))

(defmacro borra-llpc (llpc)   ; FIXME
  `(setf (llpc-id ,llpc) t))

; Funciones para el time-stamp

(defun ini-time-stamp ()
  "Inicializa el time-stamp"
  (setf time-cons-pc nil)
  (setf time-stamp 0))

(defun inc-time-stamp (cons-pc)
  "Incrementa el time-stamp si el estado time-cons-pc cambia"
  (if (not (eq time-cons-pc cons-pc))
    (setf time-cons-pc cons-pc
          time-stamp (1+ time-stamp))
    time-stamp))

; Funciones para mat-cache-pc

(defmacro cache-pc (pos)
  `(aref mat-cache-pc (first ,pos) (second ,pos)))

; Funciones para cache

(defun cache-ini (dim-tab)
  "Inicializa cache de llamadas a planta cerca"
  (setf ult-ts-lis-plantas (ini-time-stamp))
  ; Inicializa listas chaché de plantas muertas
  (setf lis-plantas-muertas nil)
  ; Inicializa matriz con llamadas a planta-cerca
  (setf mat-cache-pc (make-array dim-tab :initial-element nil)))

(defun busca-cache-pla (pla &optional (lp lis-plantas))
  "Busca ts-planta en cache de plantas recientes"
  ; FIXME Falta cond de parada bajo time-stamp
  (first (member pla lp
                 :test #'(lambda (pla ts-pla)
                           (eq-pla pla (rest ts-pla))))))

(defun cache-nueva-planta (pla)
  "Inserta nueva planta recien creada en caché"
  (if *usar-cache-pc*
      (inc-time-stamp nil))
  (push (cons time-stamp pla) lis-plantas))

(defun cache-muere-planta (pla)
  "Avisa a caché de borrado de una planta"
  ; Busca en la cache creación de dicha planta y si existe elimina
  ; Inserta en la cache de plantas muertas
  (push (cons (inc-time-stamp nil) pla) lis-plantas-muertas))

(defun cache-busca-planta-cerca (pos &optional (num 1) id)
  "Busca en la caché llamada a planta cerca"
  (let ((llpc (cache-pc pos)))
    (if (eq-llpc llpc num id)  ; Encontrada llpc anterior
        ; Revisar que alguna de las plantas de llpc no haya sido borrada
        (dolist (pdi (llpc-res llpc) (llpc-ts-res llpc))
          ; FIXME Falta cond de parada bajo time-stamp
          (if (busca-cache-pla (make-pla (first pdi) (third pdi))
                               lis-plantas-muertas)
             (progn
               (borra-llpc llpc)
               (return nil))))))) ; FIXME reutilizar tb si num son distintos

(defun cache-inserta-planta-cerca (pos id num res)
  "Inserta en la caché llamada a planta cerca y su resultado"
  (let ((llpc (cache-pc pos)))
       (if llpc
           (setf (llpc-id llpc) id
                 (llpc-num llpc) num
                 (llpc-ts llpc) time-stamp
                 (llpc-res llpc) res)
           (setf (cache-pc pos) (make-llpc id num (inc-time-stamp t) res))))
  res)

(defun cache-planta-cerca (pos &optional (num 1) id)
  "Buscar llamada a planta cerca en cache con lista limitada de plantas nuevas"
  (let* ((ts-res (cache-busca-planta-cerca pos num id))
         (ts (first ts-res))
         (lis-ord (second ts-res)))
    (if (and ts-res  ; Si se encuentra llamada igual anterior
	     (> ts ult-ts-lis-plantas)) ; y su ts esta dentro rango lis-plantas
	(do*
	    ((i 1 (1+ i))
	     (ant-lp nil lp)
	     (lp lis-plantas (rest lp))
	     (pla (rest (first lp)) (rest (first lp)))
	     (ts-pla (first (first lp)) (first (first lp))))
	    ((or (endp lp) ; FIXME reutilizar tb si num son distintos
		 (> i *lonmax-cache-pla*)
		 (< ts-pla ts))
	     (if (or (endp lp) (< ts-pla ts))
		 (subseq lis-ord 0 (min num (length lis-ord)))
	       (progn ; No hay suficientes plantas
		 (setf ult-ts-lis-plantas (first (first ant-lp)))
		 (setf (rest ant-lp) nil)
		 nil)))
	  (if (or (not id) (eq id (pla-id pla)))
	      (setf lis-ord (inserta-ord pla
					 (distancia-city pos
							 (pla-pos pla))
					 lis-ord
					 num)))))))


;;
;; Agente buscador de plantas 2 cercanas
;; -------------------------------------
;; Versión no optimizada, ideal para pocas plantas

(defun plantas-cerca2-ini (jug tipo-juego)
  "Inicializa el agente buscador de plantas cercanas 2."
  (declare (ignore jug))
  (setf lis-plantas nil)
  (if *usar-cache-pc*
    (let ((dim-tab (getf tipo-juego :dim-tab)))
      (cache-ini dim-tab)))
  t)


(defun planta-cerca2-nueva-planta (id pos)
  "Seguimiento de creación de nueva planta del agente buscador de plantas cercanas 2."
  (let ((pla (make-pla pos id)))
    (cache-nueva-planta pla)))


(defun planta-cerca2-muere-planta (id1 id2 pos)
  "Seguimiento de muerte de planta del agente buscador de plantas cercanas 2."
  (declare (ignore id2))
  (let ((pla (make-pla pos id1)))
    (if *usar-cache-pc*
      (cache-muere-planta pla))
    ; Borrarla de la cache de plantas creadas
    (setf lis-plantas (delete-if #'(lambda (ts-pla)    ; FIXME tener en cuenta ts
			             (eq-pla pla (rest ts-pla)))
			         lis-plantas
			         :count 1))))


(defun planta-cerca2 (pos &optional (num 1) id)
  "Agente buscador de plantas cercanas 2. Dada la posición POS busca las NUM plantas más cercanas a esta posición y las devuelve en una lista donde cada planta viene representada por una lista con su posición (lista con coordenada x y coordenada y), la distancia city-block a POS y su identificador ID. Si ID es distinto de NIL busca solo entre plantas generadas por el generador de plantas con identificador ID."
  (let* ((ts-res (if *usar-cache-pc*
                     (cache-busca-planta-cerca pos num id)))
         (ts (first ts-res))
         (lis-ord (second ts-res)))
    (do*
      ((lp lis-plantas (rest lp))
       (pla (rest (first lp)) (rest (first lp)))
       (ts-pla (first (first lp)) (first (first lp))))
      ((or (endp lp) ; FIXME reutilizar tb si num son distintos
           (and ts (< ts-pla ts)))
       (if *usar-cache-pc*
         (cache-inserta-planta-cerca pos id num
           (subseq lis-ord 0 (min num (length lis-ord))))
         (subseq lis-ord 0 (min num (length lis-ord)))))
      (if (or (not id) (eq id (pla-id pla)))
	  (setf lis-ord (inserta-ord pla
				     (distancia-city pos
                                                     (pla-pos pla))
				     lis-ord
				     num))))))


;;
;; Funciones de apoyo para agentes buscadores de plantas cercanas
;; --------------------------------------------------------------

(defun radio (dim)
  (+ (ceiling (1- (first dim)) 2) (ceiling (1- (second dim)) 2)))

(defun centro (dim)
  (list (1- (ceiling (first dim) 2)) (1- (ceiling (second dim) 2))))

(defmacro cluster-x (cluster)
  `(first (cluster-pos ,cluster)))

(defmacro cluster-y (cluster)
  `(second (cluster-pos ,cluster)))

(defun ajust (e d)
  (cond
   ((< e 0) 0)
   ((>= e d) (1- d))
   (t e)))

(defun ajust-cluster (cluster dim-tab)
  "Ajusta centroide de un cluster dentro del tablero"
  (let* ((x (cluster-x cluster))
	 (y (cluster-y cluster))
	 (ax (ajust x (first dim-tab)))
	 (ay (ajust y (second dim-tab)))
	 (r (cluster-rad cluster))
	 (ar (- r (abs (- ax x)) (abs (- ay y)))))
    (setf (cluster-x cluster) ax
	  (cluster-y cluster) ay
	  (cluster-rad cluster) ar)
    cluster))

(defun sub-clusters (cluster dim-tab)
  "Genera los subclusters a partir de un cluster"
  (let ((x (cluster-x cluster))
	(y (cluster-y cluster))
	(r2 (ceiling (cluster-rad cluster) 2)))
    (remove-if #'(lambda (cluster) (= (cluster-rad cluster) 0))
	       (list (ajust-cluster (make-cluster :pos (list x (- y r2)) :rad r2) dim-tab)
		     (ajust-cluster (make-cluster :pos (list (+ x r2) y) :rad r2) dim-tab)
		     (ajust-cluster (make-cluster :pos (list x (+ y r2)) :rad r2) dim-tab)
		     (ajust-cluster (make-cluster :pos (list (- x r2) y) :rad r2) dim-tab)))))

(defun hacer-arbol-clusters (lis-clusters r-max dim-tab)
  "Genera el árbol de clusters jerárquicos"
  (cond
   ((endp lis-clusters) nil)          ; Se acabaron los clusters
   ((<= (cluster-rad (first lis-clusters)) r-max) ; Cluster ha alcanzado r-max
    (cons (first lis-clusters)
	  (hacer-arbol-clusters (rest lis-clusters) r-max dim-tab)))
   (t                                 ; Hacer subclusters
    (setf (cluster-set (first lis-clusters))
	  (hacer-arbol-clusters (sub-clusters (first lis-clusters) dim-tab)
				r-max dim-tab))
    (cons (first lis-clusters)
	  (hacer-arbol-clusters (rest lis-clusters) r-max dim-tab)))))


;;
;; Sistema de producción para agente buscador de plantas cercanas
;; --------------------------------------------------------------

(defun objetivo-pc-p (nodo)
  "Indica si parar la búsqueda de plantas más cercanas"
  (if (and (listp nodo) (or (not id-pla) (eq id-pla (third nodo))))
      (progn   ; Encontrada planta, guardar y ver si terminar
	(push (list (second nodo) (first nodo) (third nodo))
	      lis-pla-cerca)
	(= (length lis-pla-cerca) num-pla))
    (null nodo)))  ; Si se acaban las plantas terminar también

(defun expande-nodo (nodo)
  "Expande un nodo, generando sus hijos"
  (if (cluster-p nodo)  ; Es un nodo cluster
      (if (not (zerop (cluster-num-pla nodo)))
	  (if (cluster-p (first (cluster-set nodo))) ; Hijos también nodos cluster
	      (mapcar #'(lambda (nodo)               ;  calculamos est. dist.
			  (setf (cluster-d-est nodo) ;   d(X,Mp) - rp
				( - (distancia-city pos-jug (cluster-pos nodo))
				    (cluster-rad nodo)))
			  nodo)
		      (cluster-set nodo))
	    (mapcar #'(lambda (planta)              ; Hijos son plantas, est. dist.
			(cons (distancia-city pos-jug (first planta)) planta))
		    (cluster-set nodo))))))         ; Nodos plantas no se expanden


(defun d-est (nodo)
  "Cálculo de la distancia estimada dese una posicion a una planta de NODO"
  (if (cluster-p nodo)
      (cluster-d-est nodo)
    (first nodo)))

(defun merge-pc-fn (new-states)
  "Fusiona la lista de hijos (NEW-STATES) a la lista de frente (GENERIC-SEARCH:*SEARCH-QUEUE*), manténiendola ordenada segun f' (GENERIC-SEARCH:*EVAL-FN*)."
  ; Ordena los estados hijos en función de f'
  (setf new-states (sort new-states #'< :key *eval-fn*))
  ; Realiza fusión ordenada de hijos en lista frente
  (fusion new-states *search-queue*))

(defun dequeue-pc-fn ()
  "Función que extrae el primer elemento de la lista de frente (GENERIC-SEARCH:*SEARCH-QUEUE*) creada con PRIORITY-A-MERGE-FN devolviendo un estado."
  (pop *search-queue*))


;;
;; Agente buscador de plantas cercanas
;; -----------------------------------
;; Adaptación de:
;; Fukunaga, Narendra (1975) "A branch and Boud Algorithm for
;; Computing k-Nearest Neighbors" IEEE Trans. on Computers, july 1975
;; Dividiendo clusters jerárquicos romboidales, con centroides (Mp) fijos
;; en el centro de cada rombo y radios máximos (rp) fijos, cada rombo
;; se subdivide en 4, hasta llegar a un radio máximo *rad-max*

(defun plantas-cerca-ini (jug tipo-juego)
  "Inicializa el agente buscador de plantas cercanas."
  (declare (ignore jug))
  (let ((dim-tab (getf tipo-juego :dim-tab)))
    (if *usar-cache-pc*
	(progn
	  (setf lis-plantas nil)
	  (cache-ini dim-tab)))
    ; Inicialización de árbol de clusters
    (setf arbol-plantas 
	  (first (hacer-arbol-clusters 
		  (list (make-cluster :pos (centro dim-tab)
				      :rad (radio dim-tab)))
		  *rad-max* dim-tab))))
  (setf dis-max (1+ (cluster-rad arbol-plantas)))
  arbol-plantas)

(defun planta-cerca-nueva-planta (id pos &optional (arbol arbol-plantas))
  "Seguimiento de creación de nueva planta del agente buscador de plantas cercanas."
  ; Inserta planta con posición POS e identificador ID en el
  ; nodo hoja donde debe ir y actualiza árbol
  ; Ojo, no se toca radio-maximo de nodos
  (if (not (cluster-p (first (cluster-set arbol)))) ; Si se ha llegado a nodo hoja
      (let ((pla (make-pla pos id)))                 ; Insertar planta
	(if *usar-cache-pc*
	    (cache-nueva-planta pla))
	(setf (cluster-num-pla arbol) (1+ (cluster-num-pla arbol)))
	(push pla (cluster-set arbol)))
    (dolist (clu (cluster-set arbol) nil)  ; sino buscar en subcluster
      (if (<= (distancia-city pos (cluster-pos clu))
              (cluster-rad clu))
         (progn
           (planta-cerca-nueva-planta id pos clu)
           (setf (cluster-num-pla arbol) (1+ (cluster-num-pla arbol)))
           (return arbol))))))

(defun planta-cerca-muere-planta (id1 id2 pos &optional (arbol arbol-plantas))
  "Seguimiento de muerte de planta del agente buscador de plantas cercanas."
  ; Borra planta con posición POS e identificador ID del
  ; nodo hoja donde debe estar y actualiza árbol
  ; Ojo, no se toca radio-maximo de nodos
  (if (not (cluster-p (first (cluster-set arbol)))) ; Si se ha llegado a nodo hoja
      (progn                    ; Borrar planta
	(if *usar-cache-pc*
	    (planta-cerca2-muere-planta id1 id2 pos))
	(setf (cluster-num-pla arbol) (1- (cluster-num-pla arbol)))
	(setf (cluster-set arbol) (delete-if #'(lambda (elm)
						 (and (equal (first elm) pos)
						      (or (not id1) (eq (second elm) id1))))
					     (cluster-set arbol)
					     :count 1)))
    (dolist (clu (cluster-set arbol) nil) ; sino buscar en subcluster
      (if (<= (distancia-city pos (cluster-pos clu))
              (cluster-rad clu))
         (progn
           (planta-cerca-muere-planta id1 id2 pos clu)
           (setf (cluster-num-pla arbol) (1- (cluster-num-pla arbol)))
           (return arbol))))))

(defun planta-cerca (pos &optional (num 1) id)
  "Dada la posición POS busca las NUM plantas más cercanas a esta posición y las devuelve en una lista donde cada planta viene representada por una lista con su posición (lista con coordenada x y coordenada y), la distancia city-block a POS y su energía. Si ID es distinto de NIL busca solo entre plantas generadas por el generador de plantas con identificador ID."
  
  (if *usar-cache-pc*
      (setf lis-pla-cerca (cache-planta-cerca pos num id))
    (setf lis-pla-cerca nil))
  (if (not lis-pla-cerca)
      (progn
	(setf pos-jug pos)
	(setf num-pla num)
	(setf id-pla id)
	(let ((saved-eval-fn *eval-fn*)
	      (saved-search-queue *search-queue*))
	  (setf *eval-fn* #'d-est)            ; Fun. f' para búsqueda
	  (setf *search-queue* nil)           ; Vaciamos lista frente
	  (pt-util:generic-search             ; Lanzamos búsqueda
	   (list arbol-plantas)
	   #'objetivo-pc-p
	   #'expande-nodo
	   :display-fn #'null
	   :merge-fn   #'merge-pc-fn
	   :dequeue-fn #'dequeue-pc-fn)
	  ; Restauramos heurística y lista frente
	  (setf *eval-fn* saved-eval-fn)
	  (setf *search-queue* saved-search-queue))
	  ; Reordenamos para devolver orden creciente en distancia
	(setf lis-pla-cerca (reverse lis-pla-cerca))))
  (if *usar-cache-pc*
      (cache-inserta-planta-cerca pos id num lis-pla-cerca)
    lis-pla-cerca))
				

;;
;; Agente buscador de plantas cercanas 3 
;; -------------------------------------
;; Hecho solo para validar versión Fukunaga & Narendra

(defun plantas-cerca3-ini (jug tipo-juego)
	(plantas-cerca-ini jug tipo-juego)
	(plantas-cerca2-ini jug tipo-juego))

(defun planta-cerca3-nueva-planta (id pos)
	(planta-cerca-nueva-planta id pos)
	(planta-cerca2-nueva-planta id pos))

(defun planta-cerca3-muere-planta (id1 id2 pos)
	(planta-cerca-muere-planta id1 id2 pos)
        (planta-cerca2-muere-planta id1 id2 pos))

(defun planta-cerca3 (pos &optional (num 1) id)
  (let ((pc1 (planta-cerca pos num id))
	(pc2 (planta-cerca2 pos num id)))
       (if (/= (second (first pc1)) (second (first pc2)))
	   (format t "Error distancias pc1: ~S, pc2: ~S~%" pc1 pc2))
       pc1))


;;
;; Variables para agente contador
;; ------------------------------

(defvar counter nil
  "Vairable contador.")

(defvar limite-cs-turno nil
  "Límite en centésimas de segundo para decidir movimiento en un turno para partida actual.")


;;
;; Funciones de apoyo para agente contador
;; ---------------------------------------

(defun ini-counter (limite-cs)
  "Inicializa el contador a LIMITE-CS centésimas de segundo."
  (setf counter
	(if (and limite-cs (> limite-cs 0))
	    (+ (get-internal-real-time)
	       (/ (* limite-cs internal-time-units-per-second)
		  100)))))

(defun test-counter (&optional hang)
  "Chequea si el contador llegado al límite, esperando si HANG es T."
  (if counter
      (do ()
	  ((> (get-internal-real-time) counter) t)
	(if (not hang)
	    (return nil)))
    hang))


;; Agente contador de tiempo
;; -------------------------

(defun contador-ini (jug tipo-juego)
  "Inicializa el agente contador de tiempo."
  (declare (ignore jug))
  (setf limite-cs-turno (getf tipo-juego :sleep-cs)))

(defun inicia-contador (&optional (limite-cs limite-cs-turno))
  "Inicializa contador del agente contador de tiempo a LIMITE-CS centésimas de segundo"
  (ini-counter limite-cs))

(defun consulta-contador (&optional hang)
  "Consulta al agente contador de tiempo si se ha cumplido el tiempo previamente inicializado con INICIA-CONTADOR, esperando si HANG es T."
  (test-counter hang))
